# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Implementation of CalcJobNode to add two numbers for testing and demonstration purposes."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import six

from aiida import orm
from aiida.common.datastructures import CalcInfo, CodeInfo
from aiida.engine import CalcJob


class ArithmeticAddCalculation(CalcJob):
    """Implementation of CalcJob to add two numbers for testing and demonstration purposes."""

    @classmethod
    def define(cls, spec):
        super(ArithmeticAddCalculation, cls).define(spec)
        spec.input('metadata.options.input_filename', valid_type=six.string_types, default='aiida.in', non_db=True)
        spec.input('metadata.options.output_filename', valid_type=six.string_types, default='aiida.out', non_db=True)
        spec.input('metadata.options.parser_name', valid_type=six.string_types, default='arithmetic.add', non_db=True)
        spec.input('x', valid_type=(orm.Int, orm.Float), help='The left operand.')
        spec.input('y', valid_type=(orm.Int, orm.Float), help='The right operand.')
        spec.output('sum', valid_type=(orm.Int, orm.Float), help='The sum of the left and right operand.')
        spec.exit_code(
            100, 'ERROR_NO_RETRIEVED_FOLDER', message='The retrieved folder data node could not be accessed.')
        spec.exit_code(
            110, 'ERROR_READING_OUTPUT_FILE', message='The output file could not be read from the retrieved folder.')
        spec.exit_code(120, 'ERROR_INVALID_OUTPUT', message='The output file contains invalid output.')

    def prepare_for_submission(self, folder):
        """
        This method is called prior to job submission with a set of calculation input nodes.
        The inputs will be validated and sanitized, after which the necessary input files will
        be written to disk in a temporary folder. A CalcInfo instance will be returned that contains
        lists of files that need to be copied to the remote machine before job submission, as well
        as file lists that are to be retrieved after job completion.

        :param folder: an aiida.common.folders.Folder to temporarily write files on disk
        :returns: CalcInfo instance
        """
        input_x = self.inputs.x
        input_y = self.inputs.y
        input_code = self.inputs.code

        self.write_input_files(folder, input_x, input_y)

        retrieve_list = self.get_retrieve_list()
        local_copy_list = self.get_local_copy_list()
        remote_copy_list = self.get_remote_copy_list()

        codeinfo = CodeInfo()
        codeinfo.cmdline_params = ['-in', self.options.input_filename]
        codeinfo.stdout_name = self.options.output_filename
        codeinfo.code_uuid = input_code.uuid

        calcinfo = CalcInfo()
        calcinfo.uuid = str(self.node.uuid)
        calcinfo.codes_info = [codeinfo]
        calcinfo.retrieve_list = retrieve_list
        calcinfo.local_copy_list = local_copy_list
        calcinfo.remote_copy_list = remote_copy_list

        return calcinfo

    def get_retrieve_list(self):
        """
        Build the list of files that are to be retrieved upon calculation completion so that they can
        be passed to the parser.

        :returns: list of resource retrieval instructions
        """
        return [self.options.output_filename]

    @staticmethod
    def get_local_copy_list():
        """
        Build the local copy list, which are files that need to be copied from the local to the remote machine

        :returns: list of resource copy instructions
        """
        return []

    @staticmethod
    def get_remote_copy_list():
        """
        Build the remote copy list, which are files that need to be copied from the remote machine from one place
        to the directory of the new calculation on the same remote machine

        :returns: list of resource copy instructions
        """
        return []

    def write_input_files(self, folder, input_x, input_y):
        """
        Take the input_parameters dictionary with the namelists and their flags
        and write the input file to disk in the temporary folder

        :param folder: an aiida.common.folders.Folder to temporarily write files on disk
        :param input_x: the numeric node representing the left operand of the summation
        :param input_y: the numeric node representing the right operand of the summation
        """
        with folder.open(self.options.input_filename, 'w', encoding='utf8') as handle:
            handle.write(u'{} {}\n'.format(input_x.value, input_y.value))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi comment` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments, options
from aiida.cmdline.utils import decorators, echo, multi_line_input
from aiida.common import exceptions


@verdi.group('comment')
def verdi_comment():
    """Inspect, create and manage node comments."""


@verdi_comment.command()
@options.NODES()
@click.argument('content', type=click.STRING, required=False)
@decorators.with_dbenv()
def add(nodes, content):
    """Add a comment to one or multiple nodes."""
    if not content:
        content = multi_line_input.edit_comment()

    for node in nodes:
        node.add_comment(content)

    echo.echo_success('comment added to {} nodes'.format(len(nodes)))


@verdi_comment.command()
@click.argument('comment_id', type=int, metavar='COMMENT_ID')
@click.argument('content', type=click.STRING, required=False)
@decorators.with_dbenv()
def update(comment_id, content):
    """Update a comment."""
    from aiida.orm.comments import Comment

    try:
        comment = Comment.objects.get(comment_id)
    except (exceptions.NotExistent, exceptions.MultipleObjectsError):
        echo.echo_critical('comment<{}> not found'.format(comment_id))

    if content is None:
        content = multi_line_input.edit_comment(comment.content)

    comment.set_content(content)

    echo.echo_success('comment<{}> updated'.format(comment_id))


@verdi_comment.command()
@options.USER()
@arguments.NODES()
@decorators.with_dbenv()
def show(user, nodes):
    """Show the comments for one or multiple nodes."""
    for node in nodes:

        all_comments = node.get_comments()

        if user is not None:
            comments = [comment for comment in all_comments if comment.user.email == user.email]

            if not comments:
                valid_users = ', '.join(set(comment.user.email for comment in all_comments))
                echo.echo_warning('no comments found for user {}'.format(user))
                echo.echo_info('valid users found for Node<{}>: {}'.format(node.pk, valid_users))

        else:
            comments = all_comments

        for comment in comments:
            comment_msg = [
                '***********************************************************',
                'Comment<{}> for Node<{}> by {}'.format(comment.id, node.pk, comment.user.email),
                'Created on {}'.format(comment.ctime.strftime('%Y-%m-%d %H:%M')),
                'Last modified on {}'.format(comment.mtime.strftime('%Y-%m-%d %H:%M')),
                '\n{}\n'.format(comment.content),
            ]
            echo.echo('\n'.join(comment_msg))

        if not comments:
            echo.echo_info('no comments found')


@verdi_comment.command()
@options.FORCE()
@click.argument('comment', type=int, required=False, metavar='COMMENT_ID')
@decorators.with_dbenv()
def remove(force, comment):
    """Remove a comment."""
    from aiida.orm.comments import Comment

    if not force:
        click.confirm('Are you sure you want to remove comment<{}>'.format(comment), abort=True)

    try:
        Comment.objects.delete(comment)
    except exceptions.NotExistent as exception:
        echo.echo_critical('failed to remove comment<{}>: {}'.format(comment, exception))
    else:
        echo.echo_success('removed comment<{}>'.format(comment))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the QueryBuilder."""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

import unittest
import warnings

from six.moves import range, zip

from aiida import orm
from aiida.backends import settings
from aiida.backends.testbase import AiidaTestCase
from aiida.common.links import LinkType

# pylint: disable=invalid-name,missing-docstring,too-many-lines


class TestQueryBuilder(AiidaTestCase):

    def setUp(self):
        super(TestQueryBuilder, self).setUp()
        self.clean_db()
        self.insert_data()

    def test_ormclass_type_classification(self):
        """
        This tests the classifications of the QueryBuilder
        """
        # pylint: disable=protected-access
        from aiida.common.exceptions import DbContentError

        qb = orm.QueryBuilder()

        # Asserting that improper declarations of the class type raise an error
        with self.assertRaises(DbContentError):
            qb._get_ormclass(None, 'data')
        with self.assertRaises(DbContentError):
            qb._get_ormclass(None, 'data.Data')
        with self.assertRaises(DbContentError):
            qb._get_ormclass(None, '.')

        # Asserting that the query type string and plugin type string are returned:
        for _cls, classifiers in (
                qb._get_ormclass(orm.StructureData, None),
                qb._get_ormclass(None, 'data.structure.StructureData.'),
        ):
            self.assertEqual(classifiers['ormclass_type_string'], orm.StructureData._plugin_type_string)  # pylint: disable=no-member

        for _cls, classifiers in (
                qb._get_ormclass(orm.Group, None),
                qb._get_ormclass(None, 'group'),
                qb._get_ormclass(None, 'Group'),
        ):
            self.assertEqual(classifiers['ormclass_type_string'], 'group')

        for _cls, classifiers in (
                qb._get_ormclass(orm.User, None),
                qb._get_ormclass(None, "user"),
                qb._get_ormclass(None, "User"),
        ):
            self.assertEqual(classifiers['ormclass_type_string'], 'user')

        for _cls, classifiers in (
                qb._get_ormclass(orm.Computer, None),
                qb._get_ormclass(None, 'computer'),
                qb._get_ormclass(None, 'Computer'),
        ):
            self.assertEqual(classifiers['ormclass_type_string'], 'computer')

        for _cls, classifiers in (
                qb._get_ormclass(orm.Data, None),
                qb._get_ormclass(None, 'data.Data.'),
        ):
            self.assertEqual(classifiers['ormclass_type_string'], orm.Data._plugin_type_string)  # pylint: disable=no-member

    def test_process_type_classification(self):
        """
        This tests the classifications of the QueryBuilder
        """
        from aiida.engine import WorkChain
        from aiida.plugins import CalculationFactory

        ArithmeticAdd = CalculationFactory('arithmetic.add')

        qb = orm.QueryBuilder()

        # pylint: disable=protected-access

        # When passing a WorkChain class, it should return the type of the corresponding Node
        # including the appropriate filter on the process_type
        _cls, classifiers = qb._get_ormclass(WorkChain, None)
        self.assertEqual(classifiers['ormclass_type_string'], 'process.workflow.workchain.WorkChainNode.')
        self.assertEqual(classifiers['process_type_string'], 'aiida.engine.processes.workchains.workchain.WorkChain')

        # When passing a WorkChainNode, no process_type filter is applied
        _cls, classifiers = qb._get_ormclass(orm.WorkChainNode, None)
        self.assertEqual(classifiers['ormclass_type_string'], 'process.workflow.workchain.WorkChainNode.')
        self.assertEqual(classifiers['process_type_string'], None)

        # Same tests for a calculation
        _cls, classifiers = qb._get_ormclass(ArithmeticAdd, None)
        self.assertEqual(classifiers['ormclass_type_string'], 'process.calculation.calcjob.CalcJobNode.')
        self.assertEqual(classifiers['process_type_string'], 'aiida.calculations:arithmetic.add')

    def test_process_query(self):
        """
        Test querying for a process class.
        """
        from aiida.engine import run, WorkChain, if_, return_, ExitCode
        from aiida.common.warnings import AiidaEntryPointWarning

        class PotentialFailureWorkChain(WorkChain):
            EXIT_STATUS = 1
            EXIT_MESSAGE = 'Well you did ask for it'
            OUTPUT_LABEL = 'optional_output'
            OUTPUT_VALUE = 144

            @classmethod
            def define(cls, spec):
                super(PotentialFailureWorkChain, cls).define(spec)
                spec.input('success', valid_type=orm.Bool)
                spec.input('through_return', valid_type=orm.Bool, default=orm.Bool(False))
                spec.input('through_exit_code', valid_type=orm.Bool, default=orm.Bool(False))
                spec.exit_code(cls.EXIT_STATUS, 'EXIT_STATUS', cls.EXIT_MESSAGE)
                spec.outline(if_(cls.should_return_out_of_outline)(return_(cls.EXIT_STATUS)), cls.failure, cls.success)
                spec.output('optional', required=False)

            def should_return_out_of_outline(self):
                return self.inputs.through_return.value

            def failure(self):
                # pylint: disable=no-else-return

                if self.inputs.success.value is False:
                    # Returning either 0 or ExitCode with non-zero status should terminate the workchain
                    if self.inputs.through_exit_code.value is False:
                        return self.EXIT_STATUS
                    else:
                        return self.exit_codes.EXIT_STATUS  # pylint: disable=no-member
                else:
                    # Returning 0 or ExitCode with zero status should *not* terminate the workchain
                    if self.inputs.through_exit_code.value is False:
                        return 0
                    else:
                        return ExitCode()

            def success(self):
                self.out(self.OUTPUT_LABEL, orm.Int(self.OUTPUT_VALUE))

        class DummyWorkChain(WorkChain):
            pass

        # Run a simple test WorkChain
        _result = run(PotentialFailureWorkChain, success=orm.Bool(True))

        # Query for nodes associated with this type of WorkChain
        qb = orm.QueryBuilder()

        with warnings.catch_warnings(record=True) as w:  # pylint: disable=no-member
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")  # pylint: disable=no-member

            qb.append(PotentialFailureWorkChain)

            # Verify some things
            assert len(w) == 1
            assert issubclass(w[-1].category, AiidaEntryPointWarning)

        # There should be one result of type WorkChainNode
        self.assertEqual(qb.count(), 1)
        self.assertTrue(isinstance(qb.all()[0][0], orm.WorkChainNode))

        # Query for nodes of a different type of WorkChain
        qb = orm.QueryBuilder()

        with warnings.catch_warnings(record=True) as w:  # pylint: disable=no-member
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")  # pylint: disable=no-member

            qb.append(DummyWorkChain)

            # Verify some things
            assert len(w) == 1
            assert issubclass(w[-1].category, AiidaEntryPointWarning)

        # There should be no result
        self.assertEqual(qb.count(), 0)

        # Query for all WorkChain nodes
        qb = orm.QueryBuilder()
        qb.append(WorkChain)

        # There should be one result
        self.assertEqual(qb.count(), 1)

    def test_simple_query_1(self):
        """
        Testing a simple query
        """
        # pylint: disable=too-many-statements

        n1 = orm.Data()
        n1.label = 'node1'
        n1.set_attribute('foo', ['hello', 'goodbye'])
        n1.store()

        n2 = orm.CalculationNode()
        n2.label = 'node2'
        n2.set_attribute('foo', 1)
        n2.store()

        n3 = orm.Data()
        n3.label = 'node3'
        n3.set_attribute('foo', 1.0000)  # Stored as fval
        n3.store()

        n4 = orm.CalculationNode()
        n4.label = 'node4'
        n4.set_attribute('foo', 'bar')
        n4.store()

        n5 = orm.Data()
        n5.label = 'node5'
        n5.set_attribute('foo', None)
        n5.store()

        n2.add_incoming(n1, link_type=LinkType.INPUT_CALC, link_label='link1')
        n3.add_incoming(n2, link_type=LinkType.CREATE, link_label='link2')

        n4.add_incoming(n3, link_type=LinkType.INPUT_CALC, link_label='link3')
        n5.add_incoming(n4, link_type=LinkType.CREATE, link_label='link4')

        qb1 = orm.QueryBuilder()
        qb1.append(orm.Node, filters={'attributes.foo': 1.000})

        self.assertEqual(len(qb1.all()), 2)

        qb2 = orm.QueryBuilder()
        qb2.append(orm.Data)
        self.assertEqual(qb2.count(), 3)

        qb2 = orm.QueryBuilder()
        qb2.append(entity_type='data.Data.')
        self.assertEqual(qb2.count(), 3)

        qb3 = orm.QueryBuilder()
        qb3.append(orm.Node, project='label', tag='node1')
        qb3.append(orm.Node, project='label', tag='node2')
        self.assertEqual(qb3.count(), 4)

        qb4 = orm.QueryBuilder()
        qb4.append(orm.CalculationNode, tag='node1')
        qb4.append(orm.Data, tag='node2')
        self.assertEqual(qb4.count(), 2)

        qb5 = orm.QueryBuilder()
        qb5.append(orm.Data, tag='node1')
        qb5.append(orm.CalculationNode, tag='node2')
        self.assertEqual(qb5.count(), 2)

        qb6 = orm.QueryBuilder()
        qb6.append(orm.Data, tag='node1')
        qb6.append(orm.Data, tag='node2')
        self.assertEqual(qb6.count(), 0)

    def test_simple_query_2(self):
        from datetime import datetime
        from aiida.common.exceptions import MultipleObjectsError, NotExistent
        n0 = orm.Data()
        n0.label = 'hello'
        n0.description = ''
        n0.set_attribute('foo', 'bar')

        n1 = orm.CalculationNode()
        n1.label = 'foo'
        n1.description = 'I am FoO'

        n2 = orm.Data()
        n2.label = 'bar'
        n2.description = 'I am BaR'

        n2.add_incoming(n1, link_type=LinkType.CREATE, link_label='random_2')
        n1.add_incoming(n0, link_type=LinkType.INPUT_CALC, link_label='random_1')

        for n in (n0, n1, n2):
            n.store()

        qb1 = orm.QueryBuilder()
        qb1.append(orm.Node, filters={'label': 'hello'})
        self.assertEqual(len(list(qb1.all())), 1)

        qh = {
            'path': [{
                'cls': orm.Node,
                'tag': 'n1'
            }, {
                'cls': orm.Node,
                'tag': 'n2',
                'with_incoming': 'n1'
            }],
            'filters': {
                'n1': {
                    'label': {
                        'ilike': '%foO%'
                    },
                },
                'n2': {
                    'label': {
                        'ilike': 'bar%'
                    },
                }
            },
            'project': {
                'n1': ['id', 'uuid', 'ctime', 'label'],
                'n2': ['id', 'description', 'label'],
            }
        }

        qb2 = orm.QueryBuilder(**qh)

        resdict = qb2.dict()
        self.assertEqual(len(resdict), 1)
        self.assertTrue(isinstance(resdict[0]['n1']['ctime'], datetime))

        res_one = qb2.one()
        self.assertTrue('bar' in res_one)

        qh = {
            'path': [{
                'cls': orm.Node,
                'tag': 'n1'
            }, {
                'cls': orm.Node,
                'tag': 'n2',
                'with_incoming': 'n1'
            }],
            'filters': {
                'n1--n2': {
                    'label': {
                        'like': '%_2'
                    }
                }
            }
        }
        qb = orm.QueryBuilder(**qh)
        self.assertEqual(qb.count(), 1)

        # Test the hashing:
        query1 = qb.get_query()
        qb.add_filter('n2', {'label': 'nonexistentlabel'})
        self.assertEqual(qb.count(), 0)

        with self.assertRaises(NotExistent):
            qb.one()
        with self.assertRaises(MultipleObjectsError):
            orm.QueryBuilder().append(orm.Node).one()

        query2 = qb.get_query()
        query3 = qb.get_query()

        self.assertTrue(id(query1) != id(query2))
        self.assertTrue(id(query2) == id(query3))

    def test_operators_eq_lt_gt(self):
        nodes = [orm.Data() for _ in range(8)]

        nodes[0].set_attribute('fa', 1)
        nodes[1].set_attribute('fa', 1.0)
        nodes[2].set_attribute('fa', 1.01)
        nodes[3].set_attribute('fa', 1.02)
        nodes[4].set_attribute('fa', 1.03)
        nodes[5].set_attribute('fa', 1.04)
        nodes[6].set_attribute('fa', 1.05)
        nodes[7].set_attribute('fa', 1.06)

        for n in nodes:
            n.store()

        self.assertEqual(orm.QueryBuilder().append(orm.Node, filters={'attributes.fa': {'<': 1}}).count(), 0)
        self.assertEqual(orm.QueryBuilder().append(orm.Node, filters={'attributes.fa': {'==': 1}}).count(), 2)
        self.assertEqual(orm.QueryBuilder().append(orm.Node, filters={'attributes.fa': {'<': 1.02}}).count(), 3)
        self.assertEqual(orm.QueryBuilder().append(orm.Node, filters={'attributes.fa': {'<=': 1.02}}).count(), 4)
        self.assertEqual(orm.QueryBuilder().append(orm.Node, filters={'attributes.fa': {'>': 1.02}}).count(), 4)
        self.assertEqual(orm.QueryBuilder().append(orm.Node, filters={'attributes.fa': {'>=': 1.02}}).count(), 5)

    def test_subclassing(self):
        s = orm.StructureData()
        s.set_attribute('cat', 'miau')
        s.store()

        d = orm.Data()
        d.set_attribute('cat', 'miau')
        d.store()

        p = orm.Dict(dict=dict(cat='miau'))
        p.store()

        # Now when asking for a node with attr.cat==miau, I want 3 esults:
        qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.cat': 'miau'})
        self.assertEqual(qb.count(), 3)

        qb = orm.QueryBuilder().append(orm.Data, filters={'attributes.cat': 'miau'})
        self.assertEqual(qb.count(), 3)

        # If I'm asking for the specific lowest subclass, I want one result
        for cls in (orm.StructureData, orm.Dict):
            qb = orm.QueryBuilder().append(cls, filters={'attributes.cat': 'miau'})
            self.assertEqual(qb.count(), 1)

        # Now I am not allow the subclassing, which should give 1 result for each
        for cls, count in ((orm.StructureData, 1), (orm.Dict, 1), (orm.Data, 1), (orm.Node, 0)):
            qb = orm.QueryBuilder().append(cls, filters={'attributes.cat': 'miau'}, subclassing=False)
            self.assertEqual(qb.count(), count)

        # Now I am testing the subclassing with tuples:
        qb = orm.QueryBuilder().append(cls=(orm.StructureData, orm.Dict), filters={'attributes.cat': 'miau'})
        self.assertEqual(qb.count(), 2)
        qb = orm.QueryBuilder().append(
            entity_type=('data.structure.StructureData.', 'data.dict.Dict.'), filters={'attributes.cat': 'miau'})
        self.assertEqual(qb.count(), 2)
        qb = orm.QueryBuilder().append(
            cls=(orm.StructureData, orm.Dict), filters={'attributes.cat': 'miau'}, subclassing=False)
        self.assertEqual(qb.count(), 2)
        qb = orm.QueryBuilder().append(
            cls=(orm.StructureData, orm.Data),
            filters={'attributes.cat': 'miau'},
        )
        self.assertEqual(qb.count(), 3)
        qb = orm.QueryBuilder().append(
            entity_type=('data.structure.StructureData.', 'data.dict.Dict.'),
            filters={'attributes.cat': 'miau'},
            subclassing=False)
        self.assertEqual(qb.count(), 2)
        qb = orm.QueryBuilder().append(
            entity_type=('data.structure.StructureData.', 'data.Data.'),
            filters={'attributes.cat': 'miau'},
            subclassing=False)
        self.assertEqual(qb.count(), 2)

    def test_list_behavior(self):
        for _i in range(4):
            orm.Data().store()

        self.assertEqual(len(orm.QueryBuilder().append(orm.Node).all()), 4)
        self.assertEqual(len(orm.QueryBuilder().append(orm.Node, project='*').all()), 4)
        self.assertEqual(len(orm.QueryBuilder().append(orm.Node, project=['*', 'id']).all()), 4)
        self.assertEqual(len(orm.QueryBuilder().append(orm.Node, project=['id']).all()), 4)
        self.assertEqual(len(orm.QueryBuilder().append(orm.Node).dict()), 4)
        self.assertEqual(len(orm.QueryBuilder().append(orm.Node, project='*').dict()), 4)
        self.assertEqual(len(orm.QueryBuilder().append(orm.Node, project=['*', 'id']).dict()), 4)
        self.assertEqual(len(orm.QueryBuilder().append(orm.Node, project=['id']).dict()), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node).iterall())), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node, project='*').iterall())), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node, project=['*', 'id']).iterall())), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node, project=['id']).iterall())), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node).iterdict())), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node, project='*').iterdict())), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node, project=['*', 'id']).iterdict())), 4)
        self.assertEqual(len(list(orm.QueryBuilder().append(orm.Node, project=['id']).iterdict())), 4)

    def test_append_validation(self):
        from aiida.common.exceptions import InputValidationError

        # So here I am giving two times the same tag
        with self.assertRaises(InputValidationError):
            orm.QueryBuilder().append(orm.StructureData, tag='n').append(orm.StructureData, tag='n')
        # here I am giving a wrong filter specifications
        with self.assertRaises(InputValidationError):
            orm.QueryBuilder().append(orm.StructureData, filters=['jajjsd'])
        # here I am giving a nonsensical projection:
        with self.assertRaises(InputValidationError):
            orm.QueryBuilder().append(orm.StructureData, project=True)

        # here I am giving a nonsensical projection for the edge:
        with self.assertRaises(InputValidationError):
            orm.QueryBuilder().append(orm.ProcessNode).append(orm.StructureData, edge_tag='t').add_projection('t', True)
        # Giving a nonsensical limit
        with self.assertRaises(InputValidationError):
            orm.QueryBuilder().append(orm.ProcessNode).limit(2.3)
        # Giving a nonsensical offset
        with self.assertRaises(InputValidationError):
            orm.QueryBuilder(offset=2.3)

        # So, I mess up one append, I want the QueryBuilder to clean it!
        with self.assertRaises(InputValidationError):
            qb = orm.QueryBuilder()
            # This also checks if we correctly raise for wrong keywords
            qb.append(orm.StructureData, tag='s', randomkeyword={})

        # Now I'm checking whether this keyword appears anywhere in the internal dictionaries:
        # pylint: disable=protected-access
        self.assertTrue('s' not in qb._projections)
        self.assertTrue('s' not in qb._filters)
        self.assertTrue('s' not in qb._tag_to_alias_map)
        self.assertTrue(len(qb._path) == 0)
        self.assertTrue(orm.StructureData not in qb._cls_to_tag_map)
        # So this should work now:
        qb.append(orm.StructureData, tag='s').limit(2).dict()

    def test_tags(self):
        qb = orm.QueryBuilder()
        qb.append(orm.Node, tag='n1')
        qb.append(orm.Node, tag='n2', edge_tag='e1', with_incoming='n1')
        qb.append(orm.Node, tag='n3', edge_tag='e2', with_incoming='n2')
        qb.append(orm.Computer, with_node='n3', tag='c1', edge_tag='nonsense')
        self.assertEqual(qb.get_used_tags(), ['n1', 'n2', 'e1', 'n3', 'e2', 'c1', 'nonsense'])

        # Now I am testing the default tags,
        qb = orm.QueryBuilder().append(orm.StructureData).append(orm.ProcessNode).append(orm.StructureData).append(
            orm.Dict, with_outgoing=orm.ProcessNode)
        self.assertEqual(qb.get_used_tags(), [
            'StructureData_1', 'ProcessNode_1', 'StructureData_1--ProcessNode_1', 'StructureData_2',
            'ProcessNode_1--StructureData_2', 'Dict_1', 'ProcessNode_1--Dict_1'
        ])
        self.assertEqual(
            qb.get_used_tags(edges=False), [
                'StructureData_1',
                'ProcessNode_1',
                'StructureData_2',
                'Dict_1',
            ])
        self.assertEqual(
            qb.get_used_tags(vertices=False),
            ['StructureData_1--ProcessNode_1', 'ProcessNode_1--StructureData_2', 'ProcessNode_1--Dict_1'])
        self.assertEqual(
            qb.get_used_tags(edges=False), [
                'StructureData_1',
                'ProcessNode_1',
                'StructureData_2',
                'Dict_1',
            ])
        self.assertEqual(
            qb.get_used_tags(vertices=False),
            ['StructureData_1--ProcessNode_1', 'ProcessNode_1--StructureData_2', 'ProcessNode_1--Dict_1'])


class TestQueryHelp(AiidaTestCase):

    def test_queryhelp(self):
        """
        Here I test the queryhelp by seeing whether results are the same as using the append method.
        I also check passing of tuples.
        """
        g = orm.Group(label='helloworld').store()
        for cls in (orm.StructureData, orm.Dict, orm.Data):
            obj = cls()
            obj.set_attribute('foo-qh2', 'bar')
            obj.store()
            g.add_nodes(obj)

        for cls, expected_count, subclassing in (
            (orm.StructureData, 1, True),
            (orm.Dict, 1, True),
            (orm.Data, 3, True),
            (orm.Data, 1, False),
            ((orm.Dict, orm.StructureData), 2, True),
            ((orm.Dict, orm.StructureData), 2, False),
            ((orm.Dict, orm.Data), 2, False),
            ((orm.Dict, orm.Data), 3, True),
            ((orm.Dict, orm.Data, orm.StructureData), 3, False),
        ):
            qb = orm.QueryBuilder()
            qb.append(cls, filters={'attributes.foo-qh2': 'bar'}, subclassing=subclassing, project='uuid')
            self.assertEqual(qb.count(), expected_count)

            qh = qb.get_json_compatible_queryhelp()
            qb_new = orm.QueryBuilder(**qh)
            self.assertEqual(qb_new.count(), expected_count)
            self.assertEqual(sorted([uuid for uuid, in qb.all()]), sorted([uuid for uuid, in qb_new.all()]))

        qb = orm.QueryBuilder().append(orm.Group, filters={'label': 'helloworld'})
        self.assertEqual(qb.count(), 1)

        qb = orm.QueryBuilder().append((orm.Group,), filters={'label': 'helloworld'})
        self.assertEqual(qb.count(), 1)

        qb = orm.QueryBuilder().append(orm.Computer,)
        self.assertEqual(qb.count(), 1)

        qb = orm.QueryBuilder().append(cls=(orm.Computer,))
        self.assertEqual(qb.count(), 1)


class TestQueryBuilderCornerCases(AiidaTestCase):
    """
    In this class corner cases of QueryBuilder are added.
    """

    def test_computer_json(self):  # pylint: disable=no-self-use
        """
        In this test we check the correct behavior of QueryBuilder when
        retrieving the _metadata and the transport_params with no content.
        Note that they are in JSON format in both backends. Forcing the
        decoding of a None value leads to an exception (this was the case
        under Django).
        """
        n1 = orm.CalculationNode()
        n1.label = 'node2'
        n1.set_attribute('foo', 1)
        n1.store()

        # Checking the correct retrieval of transport_params which is
        # a JSON field (in both backends).
        qb = orm.QueryBuilder()
        qb.append(orm.CalculationNode, project=['id'], tag='calc')
        qb.append(orm.Computer, project=['id', 'transport_params'], outerjoin=True, with_node='calc')
        qb.all()

        # Checking the correct retrieval of _metadata which is
        # a JSON field (in both backends).
        qb = orm.QueryBuilder()
        qb.append(orm.CalculationNode, project=['id'], tag='calc')
        qb.append(orm.Computer, project=['id', '_metadata'], outerjoin=True, with_node='calc')
        qb.all()


class TestAttributes(AiidaTestCase):

    def test_attribute_existence(self):
        # I'm storing a value under key whatever:
        val = 1.
        res_uuids = set()
        n1 = orm.Data()
        n1.set_attribute("whatever", 3.)
        n1.set_attribute("test_case", "test_attribute_existence")
        n1.store()

        # I want all the nodes where whatever is smaller than 1. or there is no such value:

        qb = orm.QueryBuilder()
        qb.append(
            orm.Data,
            filters={
                'or': [{
                    'attributes': {
                        '!has_key': 'whatever'
                    }
                }, {
                    'attributes.whatever': {
                        '<': val
                    }
                }],
            },
            project='uuid')
        res_query = {str(_[0]) for _ in qb.all()}
        self.assertEqual(res_query, res_uuids)

    def test_attribute_type(self):
        key = 'value_test_attr_type'
        n_int, n_float, n_str, n_str2, n_bool, n_arr = [orm.Data() for _ in range(6)]
        n_int.set_attribute(key, 1)
        n_float.set_attribute(key, 1.0)
        n_bool.set_attribute(key, True)
        n_str.set_attribute(key, '1')
        n_str2.set_attribute(key, 'one')
        n_arr.set_attribute(key, [4, 3, 5])

        for n in (n_str2, n_str, n_int, n_float, n_bool, n_arr):
            n.store()

        # Here I am testing which values contain a number 1.
        # Both 1 and 1.0 are legitimate values if ask for either 1 or 1.0
        for val in (1.0, 1):
            qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.{}'.format(key): val}, project='uuid')
            res = [str(_) for _, in qb.all()]
            self.assertEqual(set(res), set((n_float.uuid, n_int.uuid)))
            qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.{}'.format(key): {'>': 0.5}}, project='uuid')
            res = [str(_) for _, in qb.all()]
            self.assertEqual(set(res), set((n_float.uuid, n_int.uuid)))
            qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.{}'.format(key): {'<': 1.5}}, project='uuid')
            res = [str(_) for _, in qb.all()]
            self.assertEqual(set(res), set((n_float.uuid, n_int.uuid)))
        # Now I am testing the boolean value:
        qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.{}'.format(key): True}, project='uuid')
        res = [str(_) for _, in qb.all()]
        self.assertEqual(set(res), set((n_bool.uuid,)))

        qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.{}'.format(key): {'like': '%n%'}}, project='uuid')
        res = [str(_) for _, in qb.all()]
        self.assertEqual(set(res), set((n_str2.uuid,)))
        qb = orm.QueryBuilder().append(
            orm.Node, filters={'attributes.{}'.format(key): {
                                   'ilike': 'On%'
                               }}, project='uuid')
        res = [str(_) for _, in qb.all()]
        self.assertEqual(set(res), set((n_str2.uuid,)))
        qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.{}'.format(key): {'like': '1'}}, project='uuid')
        res = [str(_) for _, in qb.all()]
        self.assertEqual(set(res), set((n_str.uuid,)))
        qb = orm.QueryBuilder().append(orm.Node, filters={'attributes.{}'.format(key): {'==': '1'}}, project='uuid')
        res = [str(_) for _, in qb.all()]
        self.assertEqual(set(res), set((n_str.uuid,)))
        if settings.BACKEND == u'sqlalchemy':
            # I can't query the length of an array with Django,
            # so I exclude. Not the nicest way, But I would like to keep this piece
            # of code because of the initialization part, that would need to be
            # duplicated or wrapped otherwise.
            qb = orm.QueryBuilder().append(
                orm.Node, filters={'attributes.{}'.format(key): {
                                       'of_length': 3
                                   }}, project='uuid')
            res = [str(_) for _, in qb.all()]
            self.assertEqual(set(res), set((n_arr.uuid,)))


class QueryBuilderDateTimeAttribute(AiidaTestCase):

    @unittest.skipIf(settings.BACKEND == u'sqlalchemy', "SQLA doesn't have full datetime support in attributes")
    def test_date(self):
        from aiida.common import timezone
        from datetime import timedelta
        n = orm.Data()
        now = timezone.now()
        n.set_attribute('now', now)
        n.store()

        qb = orm.QueryBuilder().append(
            orm.Node,
            filters={
                'attributes.now': {
                    "and": [
                        {
                            ">": now - timedelta(seconds=1)
                        },
                        {
                            "<": now + timedelta(seconds=1)
                        },
                    ]
                }
            })
        self.assertEqual(qb.count(), 1)


class QueryBuilderLimitOffsetsTest(AiidaTestCase):

    def test_ordering_limits_offsets_of_results_general(self):
        # Creating 10 nodes with an attribute that can be ordered
        for i in range(10):
            n = orm.Data()
            n.set_attribute('foo', i)
            n.store()

        qb = orm.QueryBuilder().append(orm.Node, project='attributes.foo').order_by({orm.Node: 'ctime'})

        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(10)))

        # Now applying an offset:
        qb.offset(5)
        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(5, 10)))

        # Now also applying a limit:
        qb.limit(3)
        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(5, 8)))

        # Specifying the order  explicitly the order:
        qb = orm.QueryBuilder().append(
            orm.Node, project='attributes.foo').order_by({orm.Node: {
                'ctime': {
                    'order': 'asc'
                }
            }})

        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(10)))

        # Now applying an offset:
        qb.offset(5)
        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(5, 10)))

        # Now also applying a limit:
        qb.limit(3)
        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(5, 8)))

        # Reversing the order:
        qb = orm.QueryBuilder().append(
            orm.Node, project='attributes.foo').order_by({orm.Node: {
                'ctime': {
                    'order': 'desc'
                }
            }})

        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(9, -1, -1)))

        # Now applying an offset:
        qb.offset(5)
        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(4, -1, -1)))

        # Now also applying a limit:
        qb.limit(3)
        res = next(zip(*qb.all()))
        self.assertEqual(res, tuple(range(4, 1, -1)))


class QueryBuilderJoinsTests(AiidaTestCase):

    def test_joins1(self):
        # Creating n1, who will be a parent:
        parent = orm.Data()
        parent.label = 'mother'

        good_child = orm.CalculationNode()
        good_child.label = 'good_child'
        good_child.set_attribute('is_good', True)

        bad_child = orm.CalculationNode()
        bad_child.label = 'bad_child'
        bad_child.set_attribute('is_good', False)

        unrelated = orm.CalculationNode()
        unrelated.label = 'unrelated'

        for n in (good_child, bad_child, parent, unrelated):
            n.store()

        good_child.add_incoming(parent, link_type=LinkType.INPUT_CALC, link_label='parent')
        bad_child.add_incoming(parent, link_type=LinkType.INPUT_CALC, link_label='parent')

        # Using a standard inner join
        qb = orm.QueryBuilder()
        qb.append(orm.Node, tag='parent')
        qb.append(orm.Node, tag='children', project='label', filters={'attributes.is_good': True})
        self.assertEqual(qb.count(), 1)

        qb = orm.QueryBuilder()
        qb.append(orm.Node, tag='parent')
        qb.append(orm.Node, tag='children', outerjoin=True, project='label', filters={'attributes.is_good': True})
        self.assertEqual(qb.count(), 1)

    def test_joins2(self):
        # Creating n1, who will be a parent:

        students = [orm.Data() for i in range(10)]
        advisors = [orm.CalculationNode() for i in range(3)]
        for i, a in enumerate(advisors):
            a.label = 'advisor {}'.format(i)
            a.set_attribute('advisor_id', i)

        for n in advisors + students:
            n.store()

        # advisor 0 get student 0, 1
        for i in (0, 1):
            students[i].add_incoming(advisors[0], link_type=LinkType.CREATE, link_label='is_advisor_{}'.format(i))

        # advisor 1 get student 3, 4
        for i in (3, 4):
            students[i].add_incoming(advisors[1], link_type=LinkType.CREATE, link_label='is_advisor_{}'.format(i))

        # advisor 2 get student 5, 6, 7
        for i in (5, 6, 7):
            students[i].add_incoming(advisors[2], link_type=LinkType.CREATE, link_label='is_advisor_{}'.format(i))

        # let's add a differnt relationship than advisor:
        students[9].add_incoming(advisors[2], link_type=LinkType.CREATE, link_label='lover')

        self.assertEqual(
            orm.QueryBuilder().append(orm.Node).append(
                orm.Node, edge_filters={
                    'label': {
                        'like': 'is\\_advisor\\_%'
                    }
                }, tag='student').count(), 7)

        for adv_id, number_students in zip(list(range(3)), (2, 2, 3)):
            self.assertEqual(
                orm.QueryBuilder().append(orm.Node, filters={
                    'attributes.advisor_id': adv_id
                }).append(orm.Node, edge_filters={
                    'label': {
                        'like': 'is\\_advisor\\_%'
                    }
                }, tag='student').count(), number_students)

    def test_joins3_user_group(self):
        # Create another user
        new_email = "newuser@new.n"
        user = orm.User(email=new_email).store()

        # Create a group that belongs to that user
        group = orm.Group(label="node_group")
        group.user = user
        group.store()

        # Search for the group of the user
        qb = orm.QueryBuilder()
        qb.append(orm.User, tag='user', filters={'id': {'==': user.id}})
        qb.append(orm.Group, with_user='user', filters={'id': {'==': group.id}})
        self.assertEqual(qb.count(), 1, "The expected group that belongs to " "the selected user was not found.")

        # Search for the user that owns a group
        qb = orm.QueryBuilder()
        qb.append(orm.Group, tag='group', filters={'id': {'==': group.id}})
        qb.append(orm.User, with_group='group', filters={'id': {'==': user.id}})

        self.assertEqual(qb.count(), 1, "The expected user that owns the " "selected group was not found.")


class QueryBuilderPath(AiidaTestCase):

    def test_query_path(self):
        # pylint: disable=too-many-statements

        q = self.backend.query_manager
        n1 = orm.Data()
        n1.label = 'n1'
        n1.store()
        n2 = orm.CalculationNode()
        n2.label = 'n2'
        n2.store()
        n3 = orm.Data()
        n3.label = 'n3'
        n3.store()
        n4 = orm.Data()
        n4.label = 'n4'
        n4.store()
        n5 = orm.CalculationNode()
        n5.label = 'n5'
        n5.store()
        n6 = orm.Data()
        n6.label = 'n6'
        n6.store()
        n7 = orm.CalculationNode()
        n7.label = 'n7'
        n7.store()
        n8 = orm.Data()
        n8.label = 'n8'
        n8.store()
        n9 = orm.Data()
        n9.label = 'n9'
        n9.store()

        # I create a strange graph, inserting links in a order
        # such that I often have to create the transitive closure
        # between two graphs
        n3.add_incoming(n2, link_type=LinkType.CREATE, link_label='link1')
        n2.add_incoming(n1, link_type=LinkType.INPUT_CALC, link_label='link2')
        n5.add_incoming(n3, link_type=LinkType.INPUT_CALC, link_label='link3')
        n5.add_incoming(n4, link_type=LinkType.INPUT_CALC, link_label='link4')
        n4.add_incoming(n2, link_type=LinkType.CREATE, link_label='link5')
        n7.add_incoming(n6, link_type=LinkType.INPUT_CALC, link_label='link6')
        n8.add_incoming(n7, link_type=LinkType.CREATE, link_label='link7')

        # There are no parents to n9, checking that
        self.assertEqual(set([]), set(q.get_all_parents([n9.pk])))
        # There is one parent to n6
        self.assertEqual({(_,) for _ in (n6.pk,)}, {tuple(_) for _ in q.get_all_parents([n7.pk])})
        # There are several parents to n4
        self.assertEqual({(_.pk,) for _ in (n1, n2)}, {tuple(_) for _ in q.get_all_parents([n4.pk])})
        # There are several parents to n5
        self.assertEqual({(_.pk,) for _ in (n1, n2, n3, n4)}, {tuple(_) for _ in q.get_all_parents([n5.pk])})

        # Yet, no links from 1 to 8
        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n1.pk
            }, tag='anc').append(orm.Node, with_ancestors='anc', filters={
                'id': n8.pk
            }).count(), 0)

        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n8.pk
            }, tag='desc').append(orm.Node, with_descendants='desc', filters={
                'id': n1.pk
            }).count(), 0)

        n6.add_incoming(n5, link_type=LinkType.CREATE, link_label='link1')
        # Yet, now 2 links from 1 to 8
        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n1.pk
            }, tag='anc').append(orm.Node, with_ancestors='anc', filters={
                'id': n8.pk
            }).count(), 2)

        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n8.pk
            }, tag='desc').append(orm.Node, with_descendants='desc', filters={
                'id': n1.pk
            }).count(), 2)

        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n8.pk
            }, tag='desc').append(
                orm.Node,
                with_descendants='desc',
                filters={
                    'id': n1.pk
                },
                edge_filters={
                    'depth': {
                        '<': 6
                    }
                },
            ).count(), 2)
        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n8.pk
            }, tag='desc').append(
                orm.Node,
                with_descendants='desc',
                filters={
                    'id': n1.pk
                },
                edge_filters={
                    'depth': 5
                },
            ).count(), 2)
        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n8.pk
            }, tag='desc').append(
                orm.Node,
                with_descendants='desc',
                filters={
                    'id': n1.pk
                },
                edge_filters={
                    'depth': {
                        '<': 5
                    }
                },
            ).count(), 0)

        # TODO write a query that can filter certain paths by traversed ID # pylint: disable=fixme
        qb = orm.QueryBuilder().append(
            orm.Node,
            filters={
                'id': n8.pk
            },
            tag='desc',
        ).append(
            orm.Node, with_descendants='desc', edge_project='path', filters={'id': n1.pk})
        queried_path_set = {frozenset(p) for p, in qb.all()}

        paths_there_should_be = {
            frozenset([n1.pk, n2.pk, n3.pk, n5.pk, n6.pk, n7.pk, n8.pk]),
            frozenset([n1.pk, n2.pk, n4.pk, n5.pk, n6.pk, n7.pk, n8.pk])
        }

        self.assertTrue(queried_path_set == paths_there_should_be)

        qb = orm.QueryBuilder().append(
            orm.Node, filters={
                'id': n1.pk
            }, tag='anc').append(
                orm.Node, with_ancestors='anc', filters={'id': n8.pk}, edge_project='path')

        self.assertEqual({frozenset(p) for p, in qb.all()}, {
            frozenset([n1.pk, n2.pk, n3.pk, n5.pk, n6.pk, n7.pk, n8.pk]),
            frozenset([n1.pk, n2.pk, n4.pk, n5.pk, n6.pk, n7.pk, n8.pk])
        })

        n7.add_incoming(n9, link_type=LinkType.INPUT_CALC, link_label='link0')
        # Still two links...

        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n1.pk
            }, tag='anc').append(orm.Node, with_ancestors='anc', filters={
                'id': n8.pk
            }).count(), 2)

        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n8.pk
            }, tag='desc').append(orm.Node, with_descendants='desc', filters={
                'id': n1.pk
            }).count(), 2)
        n9.add_incoming(n5, link_type=LinkType.CREATE, link_label='link6')
        # And now there should be 4 nodes

        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n1.pk
            }, tag='anc').append(orm.Node, with_ancestors='anc', filters={
                'id': n8.pk
            }).count(), 4)

        self.assertEqual(
            orm.QueryBuilder().append(orm.Node, filters={
                'id': n8.pk
            }, tag='desc').append(orm.Node, with_descendants='desc', filters={
                'id': n1.pk
            }).count(), 4)

        qb = orm.QueryBuilder().append(
            orm.Node, filters={
                'id': n1.pk
            }, tag='anc').append(
                orm.Node, with_ancestors='anc', filters={'id': n8.pk}, edge_tag='edge')
        qb.add_projection('edge', 'depth')
        self.assertTrue(set(next(zip(*qb.all()))), set([5, 6]))
        qb.add_filter('edge', {'depth': 5})
        self.assertTrue(set(next(zip(*qb.all()))), set([5]))


class TestConsistency(AiidaTestCase):

    def test_create_node_and_query(self):
        """
        Testing whether creating nodes within a iterall iteration changes the results.
        """
        for _i in range(100):
            n = orm.Data()
            n.store()

        for idx, _item in enumerate(orm.QueryBuilder().append(orm.Node, project=['id',
                                                                                 'label']).iterall(batch_size=10)):
            if idx % 10 == 10:
                n = orm.Data()
                n.store()
        self.assertEqual(idx, 99)  # pylint: disable=undefined-loop-variable
        self.assertTrue(len(orm.QueryBuilder().append(orm.Node, project=['id', 'label']).all(batch_size=10)) > 99)

    def test_len_results(self):
        """
        Test whether the len of results matches the count returned.
        See also https://github.com/aiidateam/aiida_core/issues/1600
        SQLAlchemy has a deduplication strategy that leads to strange behavior, tested against here
        """
        parent = orm.CalculationNode().store()
        # adding 5 links going out:
        for inode in range(5):
            output_node = orm.Data().store()
            output_node.add_incoming(parent, link_type=LinkType.CREATE, link_label='link-{}'.format(inode))
        for projection in ('id', '*'):
            qb = orm.QueryBuilder()
            qb.append(orm.CalculationNode, filters={'id': parent.id}, tag='parent', project=projection)
            qb.append(orm.Data, with_incoming='parent')
            self.assertEqual(len(qb.all()), qb.count())


class TestManager(AiidaTestCase):

    def test_statistics(self):
        """
        Test if the statistics query works properly.

        I try to implement it in a way that does not depend on the past state.
        """
        from collections import defaultdict

        # pylint: disable=protected-access

        def store_and_add(n, statistics):
            n.store()
            statistics['total'] += 1
            statistics['types'][n._plugin_type_string] += 1  # pylint: disable=no-member
            statistics['ctime_by_day'][n.ctime.strftime('%Y-%m-%d')] += 1

        qmanager = self.backend.query_manager
        current_db_statistics = qmanager.get_creation_statistics()
        types = defaultdict(int)
        types.update(current_db_statistics['types'])
        ctime_by_day = defaultdict(int)
        ctime_by_day.update(current_db_statistics['ctime_by_day'])

        expected_db_statistics = {'total': current_db_statistics['total'], 'types': types, 'ctime_by_day': ctime_by_day}

        store_and_add(orm.Data(), expected_db_statistics)
        store_and_add(orm.Dict(), expected_db_statistics)
        store_and_add(orm.Dict(), expected_db_statistics)
        store_and_add(orm.CalculationNode(), expected_db_statistics)

        new_db_statistics = qmanager.get_creation_statistics()
        # I only check a few fields
        new_db_statistics = {k: v for k, v in new_db_statistics.items() if k in expected_db_statistics}

        expected_db_statistics = {
            k: dict(v) if isinstance(v, defaultdict) else v for k, v in expected_db_statistics.items()
        }

        self.assertEqual(new_db_statistics, expected_db_statistics)

    def test_statistics_default_class(self):
        """
        Test if the statistics query works properly.

        I try to implement it in a way that does not depend on the past state.
        """
        from collections import defaultdict

        def store_and_add(n, statistics):
            n.store()
            statistics['total'] += 1
            statistics['types'][n._plugin_type_string] += 1  # pylint: disable=no-member,protected-access
            statistics['ctime_by_day'][n.ctime.strftime('%Y-%m-%d')] += 1

        current_db_statistics = self.backend.query_manager.get_creation_statistics()
        types = defaultdict(int)
        types.update(current_db_statistics['types'])
        ctime_by_day = defaultdict(int)
        ctime_by_day.update(current_db_statistics['ctime_by_day'])

        expected_db_statistics = {'total': current_db_statistics['total'], 'types': types, 'ctime_by_day': ctime_by_day}

        store_and_add(orm.Data(), expected_db_statistics)
        store_and_add(orm.Dict(), expected_db_statistics)
        store_and_add(orm.Dict(), expected_db_statistics)
        store_and_add(orm.CalculationNode(), expected_db_statistics)

        new_db_statistics = self.backend.query_manager.get_creation_statistics()
        # I only check a few fields
        new_db_statistics = {k: v for k, v in new_db_statistics.items() if k in expected_db_statistics}

        expected_db_statistics = {
            k: dict(v) if isinstance(v, defaultdict) else v for k, v in expected_db_statistics.items()
        }

        self.assertEqual(new_db_statistics, expected_db_statistics)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Adding indices on the `input_id`, `output_id` and `type` column of the `DbLink` table

Revision ID: 5a49629f0d45
Revises: 5ddd24e52864
Create Date: 2019-03-04 16:38:42.249231

"""
# pylint: disable=invalid-name,no-member,import-error,no-name-in-module
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from alembic import op

# revision identifiers, used by Alembic.
revision = '5a49629f0d45'
down_revision = '5ddd24e52864'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.create_index(op.f('ix_db_dblink_input_id'), 'db_dblink', ['input_id'], unique=False)
    op.create_index(op.f('ix_db_dblink_output_id'), 'db_dblink', ['output_id'], unique=False)
    op.create_index(op.f('ix_db_dblink_type'), 'db_dblink', ['type'], unique=False)


def downgrade():
    """Migrations for the downgrade."""
    op.drop_index(op.f('ix_db_dblink_type'), table_name='db_dblink')
    op.drop_index(op.f('ix_db_dblink_output_id'), table_name='db_dblink')
    op.drop_index(op.f('ix_db_dblink_input_id'), table_name='db_dblink')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Unit tests for postgres database maintenance functionality"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import unittest
import mock

from pgtest.pgtest import PGTest

from aiida.manage.external.postgres import Postgres


def _try_connect_always_fail(**kwargs):  # pylint: disable=unused-argument
    """Always return False"""
    return False


class PostgresTest(unittest.TestCase):
    """Test the public API provided by the `Postgres` class"""

    def setUp(self):
        """Set up a temporary database cluster for testing potentially destructive operations"""
        self.pg_test = PGTest(max_connections=11)  # set to 11 to avoid https://github.com/jamesnunn/pgtest/issues/9
        self.dbuser = 'aiida'
        self.dbpass = 'password'
        self.dbname = 'aiida_db'

    def tearDown(self):
        self.pg_test.close()

    def _setup_postgres(self):
        postgres = Postgres(interactive=False, quiet=True, dbinfo=self.pg_test.dsn)
        postgres.determine_setup()
        return postgres

    def test_determine_setup_fail(self):
        postgres = Postgres(interactive=False, quiet=True, dbinfo={'port': '11111'})
        setup_success = postgres.determine_setup()
        self.assertFalse(setup_success)

    def test_determine_setup_success(self):
        postgres = self._setup_postgres()
        self.assertTrue(postgres.pg_execute)

    def test_setup_fail_callback(self):
        """Make sure `determine_setup` works despite wrong initial values in case of correct callback"""

        def correct_setup(interactive, dbinfo):  # pylint: disable=unused-argument
            return self.pg_test.dsn

        postgres = Postgres(interactive=False, quiet=True, dbinfo={'port': '11111'})
        postgres.set_setup_fail_callback(correct_setup)
        postgres.determine_setup()
        self.assertTrue(postgres.pg_execute)

    @mock.patch('aiida.manage.external.postgres._try_connect_psycopg', new=_try_connect_always_fail)
    @mock.patch('aiida.manage.external.postgres._try_subcmd')
    def test_fallback_on_subcmd(self, try_subcmd):
        """Ensure that accessing postgres via subcommand is tried if psycopg does not work."""
        self._setup_postgres()
        self.assertTrue(try_subcmd.call_count >= 1)

    def test_create_drop_db_user(self):
        """Check creating and dropping a user works"""
        postgres = self._setup_postgres()
        postgres.create_dbuser(self.dbuser, self.dbpass)
        self.assertTrue(postgres.dbuser_exists(self.dbuser))
        postgres.drop_dbuser(self.dbuser)
        self.assertFalse(postgres.dbuser_exists(self.dbuser))

    def test_create_drop_db(self):
        """Check creating & destroying a database"""
        postgres = self._setup_postgres()
        postgres.create_dbuser(self.dbuser, self.dbpass)
        postgres.create_db(self.dbuser, self.dbname)
        self.assertTrue(postgres.db_exists(self.dbname))
        postgres.drop_db(self.dbname)
        self.assertFalse(postgres.db_exists(self.dbname))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""AiiDA specific implementation of plumpy Ports and PortNamespaces for the ProcessSpec."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from plumpy import ports


class WithNonDb(object):  # pylint: disable=useless-object-inheritance
    """
    A mixin that adds support to a port to flag a that should not be stored
    in the database using the non_db=True flag.

    The mixins have to go before the main port class in the superclass order
    to make sure the mixin has the chance to strip out the non_db keyword.
    """

    def __init__(self, *args, **kwargs):
        non_db = kwargs.pop('non_db', False)
        super(WithNonDb, self).__init__(*args, **kwargs)
        self._non_db = non_db

    @property
    def non_db(self):
        return self._non_db


class WithSerialize(object):  # pylint: disable=useless-object-inheritance
    """
    A mixin that adds support for a serialization function which is automatically applied on inputs
    that are not AiiDA data types.
    """

    def __init__(self, *args, **kwargs):
        serializer = kwargs.pop('serializer', None)
        super(WithSerialize, self).__init__(*args, **kwargs)
        self._serializer = serializer

    def serialize(self, value):
        """
        Serialize the given value if it is not already a Data type and a serializer function is defined

        :param value: the value to be serialized
        :returns: a serialized version of the value or the unchanged value
        """
        from aiida.orm import Data

        if self._serializer is None or isinstance(value, Data):
            return value

        return self._serializer(value)


class InputPort(WithSerialize, WithNonDb, ports.InputPort):
    """
    Sub class of plumpy.InputPort which mixes in the WithSerialize and WithNonDb mixins to support automatic
    value serialization to database storable types and support non database storable input types as well.
    """

    def get_description(self):
        """
        Return a description of the InputPort, which will be a dictionary of its attributes

        :returns: a dictionary of the stringified InputPort attributes
        """
        description = super(InputPort, self).get_description()
        description['non_db'] = '{}'.format(self.non_db)

        return description


class CalcJobOutputPort(ports.OutputPort):
    """Sub class of plumpy.OutputPort which adds the `_pass_to_parser` attribute."""

    def __init__(self, *args, **kwargs):
        pass_to_parser = kwargs.pop('pass_to_parser', False)
        super(CalcJobOutputPort, self).__init__(*args, **kwargs)
        self._pass_to_parser = pass_to_parser

    @property
    def pass_to_parser(self):
        return self._pass_to_parser


class PortNamespace(WithNonDb, ports.PortNamespace):
    """
    Sub class of plumpy.PortNamespace which implements the serialize method to support automatic recursive
    serialization of a given mapping onto the ports of the PortNamespace.
    """

    def serialize(self, mapping):
        """
        Serialize the given mapping onto this Portnamespace. It will recursively call this function on any
        nested PortNamespace or the serialize function on any Ports.

        :param mapping: a mapping of values to be serialized
        :returns: the serialized mapping
        """
        if mapping is None:
            return None

        result = {}

        for name, value in mapping.items():
            if name in self:
                port = self[name]
                result[name] = port.serialize(value)
            else:
                result[name] = value

        return result



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Abstract BackendNode and BackendNodeCollection implementation."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import abc
import six

from . import backends

__all__ = ('BackendNode', 'BackendNodeCollection')


@six.add_metaclass(abc.ABCMeta)
class BackendNode(backends.BackendEntity):
    """Wrapper around a `DbNode` instance to set and retrieve data independent of the database implementation."""

    # pylint: disable=too-many-public-methods

    @abc.abstractmethod
    def clone(self):
        """Return an unstored clone of ourselves.

        :return: an unstored `BackendNode` with the exact same attributes and extras as self
        """

    @property
    def uuid(self):
        """Return the node UUID.

        :return: the string representation of the UUID
        :rtype: str or None
        """
        if self._dbmodel.uuid:
            return str(self._dbmodel.uuid)

        return None

    @property
    def node_type(self):
        """Return the node type.

        :return: the node type
        """
        return self._dbmodel.node_type

    @property
    def process_type(self):
        """Return the node process type.

        :return: the process type
        """
        return self._dbmodel.process_type

    @process_type.setter
    def process_type(self, value):
        """Set the process type.

        :param value: the new value to set
        """
        self._dbmodel.process_type = value
        if self.is_stored:
            self._increment_version_number()

    @property
    def label(self):
        """Return the node label.

        :return: the label
        """
        return self._dbmodel.label

    @label.setter
    def label(self, value):
        """Set the label.

        :param value: the new value to set
        """
        self._dbmodel.label = value
        if self.is_stored:
            self._increment_version_number()

    @property
    def description(self):
        """Return the node description.

        :return: the description
        """
        return self._dbmodel.description

    @description.setter
    def description(self, value):
        """Set the description.

        :param value: the new value to set
        """
        self._dbmodel.description = value
        if self.is_stored:
            self._increment_version_number()

    @abc.abstractproperty
    def computer(self):
        """Return the computer of this node.

        :return: the computer or None
        :rtype: `BackendComputer` or None
        """

    @computer.setter
    @abc.abstractmethod
    def computer(self, computer):
        """Set the computer of this node.

        :param computer: a `BackendComputer`
        """

    @abc.abstractproperty
    def user(self):
        """Return the user of this node.

        :return: the user
        :rtype: `BackendUser`
        """

    @user.setter
    @abc.abstractmethod
    def user(self, user):
        """Set the user of this node.

        :param user: a `BackendUser`
        """

    @property
    def ctime(self):
        """Return the node ctime.

        :return: the ctime
        """
        return self._dbmodel.ctime

    @property
    def mtime(self):
        """Return the node mtime.

        :return: the mtime
        """
        return self._dbmodel.mtime

    @property
    def version(self):
        """Return the node version.

        :return: the version
        """
        return self._dbmodel.nodeversion

    @property
    def public(self):
        """Return the node public attribute.

        :return: the public attribute
        """
        return self._dbmodel.public

    @property
    def attributes(self):
        """Return the attributes dictionary.

        .. note:: This will fetch all the attributes from the database which can be a heavy operation. If you only need
            the keys or some values, use the iterators `attributes_keys` and `attributes_items`, or the getters
            `get_attribute` and `get_attributes` instead.

        :return: the attributes as a dictionary
        """
        return dict(self.attributes_items())

    @abc.abstractmethod
    def get_attribute(self, key):
        """Return an attribute.

        :param key: name of the attribute
        :return: the value of the attribute
        :raises AttributeError: if the attribute does not exist
        """

    @abc.abstractmethod
    def get_attributes(self, keys):
        """Return a set of attributes.

        :param keys: names of the attributes
        :return: the values of the attributes
        :raises AttributeError: if at least one attribute does not exist
        """

    @abc.abstractmethod
    def set_attribute(self, key, value):
        """Set an attribute to the given value.

        :param key: name of the attribute
        :param value: value of the attribute
        """

    @abc.abstractmethod
    def set_attributes(self, attributes):
        """Set attributes.

        .. note:: This will override any existing attributes that are present in the new dictionary.

        :param attributes: the new attributes to set
        """

    @abc.abstractmethod
    def reset_attributes(self, attributes):
        """Reset the attributes.

        .. note:: This will completely reset any existing attributes and replace them with the new dictionary.

        :param attributes: the new attributes to set
        """

    @abc.abstractmethod
    def delete_attribute(self, key):
        """Delete an attribute.

        :param key: name of the attribute
        :raises AttributeError: if the attribute does not exist
        """

    @abc.abstractmethod
    def delete_attributes(self, keys):
        """Delete multiple attributes.

        .. note:: The implementation should guarantee that all the keys that are to be deleted actually exist or the
            entire operation should be canceled without any change and an ``AttributeError`` should be raised.

        :param keys: names of the attributes to delete
        :raises AttributeError: if at least on of the attribute does not exist
        """

    @abc.abstractmethod
    def clear_attributes(self):
        """Delete all attributes."""

    @abc.abstractmethod
    def attributes_items(self):
        """Return an iterator over the attribute items.

        :return: an iterator with attribute key value pairs
        """

    @abc.abstractmethod
    def attributes_keys(self):
        """Return an iterator over the attribute keys.

        :return: an iterator with attribute keys
        """

    @property
    def extras(self):
        """Return the extras dictionary.

        .. note:: This will fetch all the extras from the database which can be a heavy operation. If you only need
            the keys or some values, use the iterators `extras_keys` and `extras_items`, or the getters `get_extra`
            and `get_extras` instead.

        :return: the extras as a dictionary
        """
        return dict(self.extras_items())

    @abc.abstractmethod
    def get_extra(self, key):
        """Return an extra.

        :param key: name of the extra
        :return: the value of the extra
        :raises AttributeError: if the extra does not exist
        """

    @abc.abstractmethod
    def get_extras(self, keys):
        """Return a set of extras.

        :param keys: names of the extras
        :return: the values of the extras
        :raises AttributeError: if at least one extra does not exist
        """

    @abc.abstractmethod
    def set_extra(self, key, value, increase_version=True):
        """Set an extra to the given value.

        :param key: name of the extra
        :param value: value of the extra
        :param increase_version: boolean, if True will increase the node version upon successfully setting the extra
        """

    @abc.abstractmethod
    def set_extras(self, extras):
        """Set extras.

        .. note:: This will override any existing extras that are present in the new dictionary.

        :param extras: the new extras to set
        """

    @abc.abstractmethod
    def reset_extras(self, extras):
        """Reset the extras.

        .. note:: This will completely reset any existing extras and replace them with the new dictionary.

        :param extras: the new extras to set
        """

    @abc.abstractmethod
    def delete_extra(self, key):
        """Delete an extra.

        :param key: name of the extra
        :raises AttributeError: if the extra does not exist
        """

    @abc.abstractmethod
    def delete_extras(self, keys):
        """Delete multiple extras.

        .. note:: The implementation should guarantee that all the keys that are to be deleted actually exist or the
            entire operation should be canceled without any change and an ``AttributeError`` should be raised.

        :param keys: names of the extras to delete
        :raises AttributeError: if at least on of the extra does not exist
        """

    @abc.abstractmethod
    def clear_extras(self):
        """Delete all extras."""

    @abc.abstractmethod
    def extras_items(self):
        """Return an iterator over the extra items.

        :return: an iterator with extra key value pairs
        """

    @abc.abstractmethod
    def extras_keys(self):
        """Return an iterator over the attribute keys.

        :return: an iterator with attribute keys
        """

    @abc.abstractmethod
    def add_incoming(self, source, link_type, link_label):
        """Add a link of the given type from a given node to ourself.

        :param source: the node from which the link is coming
        :param link_type: the link type
        :param link_label: the link label
        :return: True if the proposed link is allowed, False otherwise
        :raise TypeError: if `source` is not a Node instance or `link_type` is not a `LinkType` enum
        :raise ValueError: if the proposed link is invalid
        """

    @abc.abstractmethod
    def store(self, attributes=None, links=None, with_transaction=True):
        """Store the node in the database.

        :param attributes: optional attributes to set before storing, will override any existing attributes
        :param links: optional links to add before storing
        :parameter with_transaction: if False, do not use a transaction because the caller will already have opened one.
        """

    @abc.abstractmethod
    def _increment_version_number(self):
        """Increment the node version number of this node by one directly in the database."""


@six.add_metaclass(abc.ABCMeta)
class BackendNodeCollection(backends.BackendCollection[BackendNode]):
    """The collection of `BackendNode` entries."""

    # pylint: disable=too-few-public-methods

    ENTITY_CLASS = BackendNode

    @abc.abstractmethod
    def delete(self, pk):
        """Remove a Node entry from the collection with the given id

        :param pk: id of the node to delete
        """



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Class and decorators to generate processes out of simple python functions."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import collections
import functools
import inspect

from six.moves import zip  # pylint: disable=unused-import

from aiida.common.lang import override
from aiida.manage.manager import get_manager

from .process import Process

__all__ = ('calcfunction', 'workfunction')


def calcfunction(function):
    """
    A decorator to turn a standard python function into a calcfunction.
    Example usage:

    >>> from aiida.orm import Int
    >>>
    >>> # Define the calcfunction
    >>> @calcfunction
    >>> def sum(a, b):
    >>>    return a + b
    >>> # Run it with some input
    >>> r = sum(Int(4), Int(5))
    >>> print(r)
    9
    >>> r.get_incoming().all() # doctest: +SKIP
    [Neighbor(link_type='', link_label='result',
    node=<CalcFunctionNode: uuid: ce0c63b3-1c84-4bb8-ba64-7b70a36adf34 (pk: 3567)>)]
    >>> r.get_incoming().get_node_by_label('result').get_incoming().all_nodes()
    [4, 5]

    """
    from aiida.orm import CalcFunctionNode
    return process_function(node_class=CalcFunctionNode)(function)


def workfunction(function):
    """
    A decorator to turn a standard python function into a workfunction.
    Example usage:

    >>> from aiida.orm import Int
    >>>
    >>> # Define the workfunction
    >>> @workfunction
    >>> def select(a, b):
    >>>    return a
    >>> # Run it with some input
    >>> r = select(Int(4), Int(5))
    >>> print(r)
    4
    >>> r.get_incoming().all() # doctest: +SKIP
    [Neighbor(link_type='', link_label='result',
    node=<WorkFunctionNode: uuid: ce0c63b3-1c84-4bb8-ba64-7b70a36adf34 (pk: 3567)>)]
    >>> r.get_incoming().get_node_by_label('result').get_incoming().all_nodes()
    [4, 5]

    """
    from aiida.orm import WorkFunctionNode
    return process_function(node_class=WorkFunctionNode)(function)


def process_function(node_class):
    """
    The base function decorator to create a FunctionProcess out of a normal python function.

    :param node_class: the ORM class to be used as the Node record for the FunctionProcess
    """

    @staticmethod
    @property
    def is_process_function():
        return True

    def decorator(function):
        """
        Turn the decorated function into a FunctionProcess.

        :param function: the actual decorated function that the FunctionProcess represents
        """
        process_class = FunctionProcess.build(function, node_class=node_class)

        def run_get_node(*args, **kwargs):
            """
            Run the FunctionProcess with the supplied inputs in a local runner.

            The function will have to create a new runner for the FunctionProcess instead of using the global runner,
            because otherwise if this workfunction were to call another one from within its scope, that would use
            the same runner and it would be blocking the event loop from continuing.

            :param args: input arguments to construct the FunctionProcess
            :param kwargs: input keyword arguments to construct the FunctionProcess
            :return: tuple of the outputs of the process and the calculation node
            """
            runner = get_manager().create_runner(with_persistence=False)
            inputs = process_class.create_inputs(*args, **kwargs)

            # Remove all the known inputs from the kwargs
            for port in process_class.spec().inputs:
                kwargs.pop(port, None)

            # If any kwargs remain, the spec should be dynamic, so we raise if it isn't
            if kwargs and not process_class.spec().inputs.dynamic:
                raise ValueError('{} does not support these kwargs: {}'.format(function.__name__, kwargs.keys()))

            process = process_class(inputs=inputs, runner=runner)
            result = process.execute()

            # Close the runner properly
            runner.close()

            store_provenance = inputs.get('metadata', {}).get('store_provenance', True)
            if not store_provenance:
                process.node._storable = False  # pylint: disable=protected-access
                process.node._unstorable_message = 'cannot store node because it was run with `store_provenance=False`'  # pylint: disable=protected-access

            return result, process.node

        @functools.wraps(function)
        def decorated_function(*args, **kwargs):
            """This wrapper function is the actual function that is called."""
            result, _ = run_get_node(*args, **kwargs)
            return result

        decorated_function.run_get_node = run_get_node
        decorated_function.is_process_function = is_process_function

        return decorated_function

    return decorator


class FunctionProcess(Process):
    """Function process class used for turning functions into a Process"""

    _func_args = None

    @staticmethod
    def _func(*_args, **_kwargs):
        """
        This is used internally to store the actual function that is being
        wrapped and will be replaced by the build method.
        """
        return {}

    @staticmethod
    def build(func, node_class):
        """
        Build a Process from the given function.

        All function arguments will be assigned as process inputs. If keyword arguments are specified then
        these will also become inputs.

        :param func: The function to build a process from
        :param node_class: Provide a custom node class to be used, has to be constructable with no arguments. It has to
            be a sub class of `ProcessNode` and the mixin :class:`~aiida.orm.utils.mixins.FunctionCalculationMixin`.
        :type node_class: :class:`aiida.orm.nodes.process.process.ProcessNode`
        :return: A Process class that represents the function
        :rtype: :class:`FunctionProcess`
        """
        from aiida import orm
        from aiida.orm import ProcessNode
        from aiida.orm.utils.mixins import FunctionCalculationMixin

        if not issubclass(node_class, ProcessNode) or not issubclass(node_class, FunctionCalculationMixin):
            raise TypeError('the node_class should be a sub class of `ProcessNode` and `FunctionCalculationMixin`')

        args, varargs, keywords, defaults = inspect.getargspec(func)  # pylint: disable=deprecated-method
        nargs = len(args)
        ndefaults = len(defaults) if defaults else 0
        first_default_pos = nargs - ndefaults

        if varargs is not None:
            raise ValueError('variadic arguments are not supported')

        def _define(cls, spec):
            """Define the spec dynamically"""
            super(FunctionProcess, cls).define(spec)

            for i, arg in enumerate(args):
                default = ()
                if i >= first_default_pos:
                    default = defaults[i - first_default_pos]

                # If the keyword was already specified, simply override the default
                if spec.has_input(arg):
                    spec.inputs[arg].default = default
                else:
                    # If the default is `None` make sure that the port also accepts a `NoneType`
                    # Note that we cannot use `None` because the validation will call `isinstance` which does not work
                    # when passing `None`, but it does work with `NoneType` which is returned by calling `type(None)`
                    if default is None:
                        valid_type = (orm.Data, type(None))
                    else:
                        valid_type = (orm.Data,)

                    spec.input(arg, valid_type=valid_type, default=default)

            # If the function support kwargs then allow dynamic inputs, otherwise disallow
            spec.inputs.dynamic = keywords is not None

            # Function processes return data types
            spec.outputs.valid_type = orm.Data

        return type(
            func.__name__, (FunctionProcess,), {
                '_func': staticmethod(func),
                Process.define.__name__: classmethod(_define),
                '_func_args': args,
                '_node_class': node_class
            })

    @classmethod
    def create_inputs(cls, *args, **kwargs):
        """Create the input args for the FunctionProcess"""
        ins = {}
        if kwargs:
            ins.update(kwargs)
        if args:
            ins.update(cls.args_to_dict(*args))
        return ins

    @classmethod
    def args_to_dict(cls, *args):
        """
        Create an input dictionary (i.e. label: value) from supplied args.

        :param args: The values to use
        :return: A label: value dictionary
        """
        return dict(list(zip(cls._func_args, args)))

    @classmethod
    def get_or_create_db_record(cls):
        return cls._node_class()

    def __init__(self, *args, **kwargs):
        if kwargs.get('enable_persistence', False):
            raise RuntimeError('Cannot persist a function process')
        super(FunctionProcess, self).__init__(enable_persistence=False, *args, **kwargs)

    @property
    def process_class(self):
        """
        Return the class that represents this Process, for the FunctionProcess this is the function itself.

        For a standard Process or sub class of Process, this is the class itself. However, for legacy reasons,
        the Process class is a wrapper around another class. This function returns that original class, i.e. the
        class that really represents what was being executed.
        """
        return self._func

    def execute(self):
        """Execute the process."""
        result = super(FunctionProcess, self).execute()

        # FunctionProcesses can return a single value as output, and not a dictionary, so we should also return that
        if len(result) == 1 and self.SINGLE_OUTPUT_LINKNAME in result:
            return result[self.SINGLE_OUTPUT_LINKNAME]

        return result

    @override
    def _setup_db_record(self):
        """Set up the database record for the process."""
        super(FunctionProcess, self)._setup_db_record()
        self.node.store_source_info(self._func)

    @override
    def run(self):
        """Run the process"""
        from aiida.orm import Data
        from .exit_code import ExitCode

        args = []

        # Split the inputs into positional and keyword arguments
        args = [None] * len(self._func_args)
        kwargs = {}
        for name, value in self.inputs.items():
            try:
                if self.spec().inputs[name].non_db:
                    # Don't consider non-database inputs
                    continue
            except KeyError:
                pass  # No port found

            # Check if it is a positional arg, if not then keyword
            try:
                args[self._func_args.index(name)] = value
            except ValueError:
                kwargs[name] = value

        result = self._func(*args, **kwargs)

        if result is None or isinstance(result, ExitCode):
            return result

        if isinstance(result, Data):
            self.out(self.SINGLE_OUTPUT_LINKNAME, result)
        elif isinstance(result, collections.Mapping):
            for name, value in result.items():
                self.out(name, value)
        else:
            raise TypeError("Function process returned an output with unsupported type '{}'\n"
                            "Must be a Data type or a mapping of {{string: Data}}".format(result.__class__))

        return ExitCode()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Reusable command line interface options for Computer commands."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from aiida.cmdline.params import options, types
from aiida.cmdline.params.options.interactive import InteractiveOption
from aiida.cmdline.params.options.overridable import OverridableOption


def should_call_default_mpiprocs_per_machine(ctx):  # pylint: disable=invalid-name
    """
    Return True if the scheduler can accept 'default_mpiprocs_per_machine',
    False otherwise.

    If there is a problem in determining the scheduler, return True to
    avoid exceptions.
    """
    from aiida.common.exceptions import ValidationError

    scheduler_ep = ctx.params['scheduler']
    if scheduler_ep is not None:
        try:
            scheduler_cls = scheduler_ep.load()
        except ImportError:
            raise ImportError("Unable to load the '{}' scheduler".format(scheduler_ep.name))
    else:
        raise ValidationError(
            "The should_call_... function should always be run (and prompted) AFTER asking for a scheduler")

    job_resource_cls = scheduler_cls.job_resource_class
    if job_resource_cls is None:
        # Odd situation...
        return False

    return job_resource_cls.accepts_default_mpiprocs_per_machine()


LABEL = options.LABEL.clone(
    prompt='Computer label', cls=InteractiveOption, required=True, type=types.NonEmptyStringParamType())

HOSTNAME = options.HOSTNAME.clone(
    prompt='Hostname',
    cls=InteractiveOption,
    required=True,
    help='The fully qualified hostname of this computer; for local transports, use localhost.')

DESCRIPTION = options.DESCRIPTION.clone(
    prompt='Description', cls=InteractiveOption, help='A human-readable description of this computer.')

ENABLED = OverridableOption(
    '-e/-d',
    '--enabled/--disabled',
    is_flag=True,
    default=True,
    help='Calculations on disabled computers will not be submitted until the computer is re-enabled.',
    prompt='Enable the computer?',
    cls=InteractiveOption,
    # IMPORTANT! Do not specify explicitly type=click.BOOL,
    # Otherwise you would not get a default value when prompting
)

TRANSPORT = options.TRANSPORT.clone(prompt='Transport plugin', cls=InteractiveOption)

SCHEDULER = options.SCHEDULER.clone(prompt='Scheduler plugin', cls=InteractiveOption)

SHEBANG = OverridableOption(
    '--shebang',
    prompt='Shebang line (first line of each script, starting with #!)',
    default='#!/bin/bash',
    cls=InteractiveOption,
    help='This line specifies the first line of the submission script for this computer.',
    type=types.ShebangParamType())

WORKDIR = OverridableOption(
    '-w',
    '--work-dir',
    prompt='Work directory on the computer',
    default='/scratch/{username}/aiida/',
    cls=InteractiveOption,
    help='The absolute path of the directory on the computer where AiiDA will '
    'run the calculations (typically, the scratch of the computer). You '
    'can use the {username} replacement, that will be replaced by your '
    'username on the remote computer.')

MPI_RUN_COMMAND = OverridableOption(
    '-m',
    '--mpirun-command',
    prompt='Mpirun command',
    default='mpirun -np {tot_num_mpiprocs}',
    cls=InteractiveOption,
    help='The mpirun command needed on the cluster to run parallel MPI '
    'programs. You can use the {tot_num_mpiprocs} replacement, that will be '
    'replaced by the total number of cpus, or the other scheduler-dependent '
    'replacement fields (see the scheduler docs for more information).',
    type=types.MpirunCommandParamType())

MPI_PROCS_PER_MACHINE = OverridableOption(
    '--mpiprocs-per-machine',
    prompt='Default number of CPUs per machine',
    cls=InteractiveOption,
    prompt_fn=should_call_default_mpiprocs_per_machine,
    required_fn=False,
    type=click.INT,
    help='Enter here the default number of MPI processes per machine (node) that '
    'should be used if nothing is otherwise specified. Pass the digit 0 '
    'if you do not want to provide a default value.',
)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utilities for dealing with links between nodes."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from collections import namedtuple

from aiida.common import exceptions

__all__ = ('LinkPair', 'LinkTriple', 'LinkManager', 'validate_link')

LinkPair = namedtuple('LinkPair', ['link_type', 'link_label'])
LinkTriple = namedtuple('LinkTriple', ['node', 'link_type', 'link_label'])


def validate_link(source, target, link_type, link_label):
    """
    Validate adding a link of the given type and label from a given node to ourself.

    This function will first validate the class types of the inputs and will subsequently validate whether a link of
    the specified type is allowed at all between the nodes types of the source and target.

    Subsequently, the validity of the "indegree" and "outdegree" of the proposed link is validated, which means
    validating that the uniqueness constraints of the incoming links into the target node and the outgoing links from
    the source node are not violated. In AiiDA's provenance graph each link type has one of the following three types
    of "degree" character::

        * unique
        * unique pair
        * unique triple

    Each degree character has a different unique constraint on its links, here defined for the indegree::

        * unique: any target node, it can only have a single incoming link of this type, regardless of the link label.
        * unique pair: a node can have an infinite amount of incoming links of this type, as long as the labels within
            that sub set, are unique. In short, it is the link pair, i.e. the tuple of the link type and label, that has
            a uniquess constraint for the incoming links to a given node.
        * unique triple: a node can have an infinite amount of incoming links of this type, as long as the triple tuple
            of source node, link type and link label is unique. In other words, it is the link triple that has a
            uniqueness constraint for the incoming links.

    The same holds for outdegree, but then it concerns outgoing links from the source node to the target node.

    For illustration purposes, consider the following example provenance graphs that are considered legal, where
    `WN`, `DN` and `CN` represent a `WorkflowNode`, a `DataNode` and a `CalculationNode`, respectively::

                    1                    2                    3
            ______     ______          ______          ______     ______
           |      |   |      |        |      |        |      |   |      |
           |  WN  |   |  DN  |        |  DN  |        |  WN  |   |  WN  |
           |______|   |______|        |______|        |______|   |______|
                |     /                |   |               |     /
              a |    / a             a |   | b           a |    / a
               _|___/                  |___|_             _|___/
              |      |                |      |           |      |
              |  CN  |                |  CN  |           |  DN  |
              |______|                |______|           |______|

    In example 1, the link uniqueness constraint is not violated because despite the labels having the same label `a`,
    their link types, `CALL_CALC` and `INPUT_CALC`, respectively, are different and their `unique_pair` indegree is
    not violated.

    Similarly, in the second example, the constraint is not violated, because despite both links having the same link
    type `INPUT_CALC`, the have different labels, so the `unique_pair` indegree of the `INPUT_CALC` is not violated.

    Finally, in the third example, we see two `WorkflowNodes` both returning the same `DataNode` and with the same
    label. Despite the two incoming links here having both the same type as well as the same label, the uniqueness
    constraint is not violated, because the indegree for `RETURN` links is `unique_triple` which means that the triple
    of source node and link type and label should be unique.

    :param source: the node from which the link is coming
    :param target: the node to which the link is going
    :param link_type: the type of link
    :param link_label: link label
    :raise TypeError: if `source` or `target` is not a Node instance, or `link_type` is not a `LinkType` enum
    :raise ValueError: if the proposed link is invalid
    """
    from aiida.common.links import LinkType
    from aiida.orm import Node, Data, CalculationNode, WorkflowNode

    if not isinstance(link_type, LinkType):
        raise TypeError('the link_type should be a value from the LinkType enum')

    if not isinstance(source, Node):
        raise TypeError('the source should be a Node instance')

    if not isinstance(target, Node):
        raise TypeError('the target should be a Node instance')

    if source.uuid is None or target.uuid is None:
        raise ValueError('source or target node does not have a UUID')

    if source.uuid == target.uuid:
        raise ValueError('cannot add a link to oneself')

    # For each link type, define a tuple that defines the valid types for the source and target node, as well as
    # the outdegree and indegree character. If the degree is `unique` that means that there can only be a single
    # link of this type regardless of the label. If instead it is `unique_label`, an infinite amount of links of that
    # type can be defined, as long as the link label is unique for the sub set of links of that type. Finally, for
    # `unique_triple` the triple of node, link type and link label has to be unique.
    link_mapping = {
        LinkType.CALL_CALC: (WorkflowNode, CalculationNode, 'unique_triple', 'unique'),
        LinkType.CALL_WORK: (WorkflowNode, WorkflowNode, 'unique_triple', 'unique'),
        LinkType.CREATE: (CalculationNode, Data, 'unique_pair', 'unique'),
        LinkType.INPUT_CALC: (Data, CalculationNode, 'unique_triple', 'unique_pair'),
        LinkType.INPUT_WORK: (Data, WorkflowNode, 'unique_triple', 'unique_pair'),
        LinkType.RETURN: (WorkflowNode, Data, 'unique_pair', 'unique_triple'),
    }

    type_source, type_target, outdegree, indegree = link_mapping[link_type]

    if not isinstance(source, type_source) or not isinstance(target, type_target):
        raise ValueError('cannot add a {} link from {} to {}'.format(link_type, type(source), type(target)))

    # Validate the outdegree of the source node
    outgoing = source.get_outgoing(link_type=link_type)

    # If the outdegree is `unique` there cannot already be any other incoming links of that type
    if outdegree == 'unique' and outgoing.all():
        raise ValueError('node<{}> already has an outgoing {} link'.format(source.uuid, link_type))

    # If the outdegree is `unique_pair` than the link labels for outgoing links of this type should be unique
    elif outdegree == 'unique_pair' and LinkPair(link_type, link_label) in outgoing.all_link_pairs():
        raise ValueError('node<{}> already has an outgoing {} link with label "{}"'.format(
            target.uuid, link_type, link_label))

    # If the outdegree is `unique_triple` than the link triples of link type, link label and target should be unique
    elif outdegree == 'unique_triple' and LinkTriple(target, link_type, link_label) in outgoing.all():
        raise ValueError('node<{}> already has an outgoing {} link with label "{}" from node<{}>'.format(
            source.uuid, link_type, link_label, target.uuid))

    # Validate the indegree of the target node
    incoming = target.get_incoming(link_type=link_type)

    # If the indegree is `unique` there cannot already be any other incoming links of that type
    if indegree == 'unique' and incoming.all():
        raise ValueError('node<{}> already has an incoming {} link'.format(target.uuid, link_type))

    # If the indegree is `unique_pair` than the link labels for incoming links of this type should be unique
    elif indegree == 'unique_pair' and LinkPair(link_type, link_label) in incoming.all_link_pairs():
        raise ValueError('node<{}> already has an incoming {} link with label "{}"'.format(
            target.uuid, link_type, link_label))

    # If the indegree is `unique_triple` than the link triples of link type, link label and source should be unique
    elif indegree == 'unique_triple' and LinkTriple(source, link_type, link_label) in incoming.all():
        raise ValueError('node<{}> already has an incoming {} link with label "{}" from node<{}>'.format(
            target.uuid, link_type, link_label, source.uuid))


class LinkManager(object):  # pylint: disable=useless-object-inheritance
    """
    Class to convert a list of LinkTriple tuples into an iterator.

    It defines convenience methods to retrieve certain subsets of LinkTriple while checking for consistency.
    For example::

        LinkManager.one(): returns the only entry in the list or it raises an exception
        LinkManager.first(): returns the first entry from the list
        LinkManager.all(): returns all entries from list

    The methods `all_nodes` and `all_link_labels` are syntactic sugar wrappers around `all` to get a list of only the
    incoming nodes or link labels, respectively.
    """

    def __init__(self, link_triples):
        """Initialise the collection."""
        self.link_triples = link_triples

    def __iter__(self):
        """Return an iterator of LinkTriple instances.

        :return: iterator of LinkTriple instances
        """
        return iter(self.link_triples)

    def __next__(self):
        """Return the next element in the iterator.

        :return: LinkTriple
        """
        for link_triple in self.link_triples:
            yield link_triple

    def next(self):
        """Return the next element in the iterator.

        :return: LinkTriple
        """
        return self.__next__()

    def one(self):
        """Return a single entry from the iterator.

        If the iterator contains no or more than one entry, an exception will be raised
        :return: LinkTriple instance
        :raises ValueError: if the iterator contains anything but one entry
        """
        if self.link_triples:
            if len(self.link_triples) > 1:
                raise ValueError('more than one entry found')
            return self.link_triples[0]

        raise ValueError('no entries found')

    def first(self):
        """Return the first entry from the iterator.

        :return: LinkTriple instance
        :raises ValueError: if the iterator contains anything but one entry
        """
        if self.link_triples:
            return self.link_triples[0]
        raise ValueError('no entries found')

    def all(self):
        """Return all entries from the list.

        :return: list of LinkTriple instances
        """
        return self.link_triples

    def all_nodes(self):
        """Return a list of all nodes.

        :return: list of nodes
        """
        return [entry.node for entry in self.all()]

    def all_link_pairs(self):
        """Return a list of all link pairs.

        :return: list of LinkPair instances
        """
        return [LinkPair(entry.link_type, entry.link_label) for entry in self.all()]

    def all_link_labels(self):
        """Return a list of all link labels.

        :return: list of link labels
        """
        return [entry.link_label for entry in self.all()]

    def get_node_by_label(self, label):
        """Return the node from list for given label.

        :return: node that corresponds to the given label
        :raises aiida.common.NotExistent: if the label is not present among the link_triples
        """
        matching_entry = None
        for entry in self.link_triples:
            if entry.link_label == label:
                if matching_entry is None:
                    matching_entry = entry.node
                else:
                    raise exceptions.MultipleObjectsError(
                        'more than one neighbor with the label {} found'.format(label))

        if matching_entry is None:
            raise exceptions.NotExistent('no neighbor with the label {} found'.format(label))

        return matching_entry



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Django implementation of `aiida.orm.implementation.backends.Backend`."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from contextlib import contextmanager

# pylint: disable=import-error,no-name-in-module
from django.db import models, transaction

from aiida.backends.djsite.queries import DjangoQueryManager
from aiida.backends.djsite.utils import migrate_database

from ..sql import SqlBackend
from . import authinfos
from . import comments
from . import computers
from . import convert
from . import groups
from . import logs
from . import nodes
from . import querybuilder
from . import users

__all__ = ('DjangoBackend',)


class DjangoBackend(SqlBackend[models.Model]):
    """Django implementation of `aiida.orm.implementation.backends.Backend`."""

    def __init__(self):
        """Construct the backend instance by initializing all the collections."""
        self._authinfos = authinfos.DjangoAuthInfoCollection(self)
        self._comments = comments.DjangoCommentCollection(self)
        self._computers = computers.DjangoComputerCollection(self)
        self._groups = groups.DjangoGroupCollection(self)
        self._logs = logs.DjangoLogCollection(self)
        self._nodes = nodes.DjangoNodeCollection(self)
        self._query_manager = DjangoQueryManager(self)
        self._users = users.DjangoUserCollection(self)

    @staticmethod
    def migrate():
        migrate_database()

    @property
    def authinfos(self):
        return self._authinfos

    @property
    def comments(self):
        return self._comments

    @property
    def computers(self):
        return self._computers

    @property
    def groups(self):
        return self._groups

    @property
    def logs(self):
        return self._logs

    @property
    def nodes(self):
        return self._nodes

    @property
    def query_manager(self):
        return self._query_manager

    def query(self):
        return querybuilder.DjangoQueryBuilder(self)

    @property
    def users(self):
        return self._users

    @staticmethod
    def transaction():
        """Open a transaction to be used as a context manager."""
        return transaction.atomic()

    # Below are abstract methods inherited from `aiida.orm.implementation.sql.backends.SqlBackend`

    def get_backend_entity(self, model):
        """Return a `BackendEntity` instance from a `DbModel` instance."""
        return convert.get_backend_entity(model, self)

    @contextmanager
    def cursor(self):
        """Return a psycopg cursor to be used in a context manager.

        :return: a psycopg cursor
        :rtype: :class:`psycopg2.extensions.cursor`
        """
        try:
            yield self.get_connection().cursor()
        finally:
            pass

    def execute_raw(self, query):
        """Execute a raw SQL statement and return the result.

        :param query: a string containing a raw SQL statement
        :return: the result of the query
        """
        with self.cursor() as cursor:
            cursor.execute(query)
            results = cursor.fetchall()

        return results

    @staticmethod
    def get_connection():
        """
        Get the Django connection

        :return: the django connection
        """
        # pylint: disable=import-error,no-name-in-module
        from django.db import connection
        # For now we just return the global but if we ever support multiple Django backends
        # being loaded this should be specific to this backend
        return connection



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for authinfo backend classes."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import abc
import six

from . import backends

__all__ = ('BackendAuthInfo', 'BackendAuthInfoCollection')


@six.add_metaclass(abc.ABCMeta)
class BackendAuthInfo(backends.BackendEntity):
    """
    Base class for backend authorization information which contains computer configuration
    specific to a given user (authorization info and other metadata, like
    how often to check on a given computer etc.)
    """

    METADATA_WORKDIR = 'workdir'

    def pk(self):
        """
        Return the principal key in the DB.
        """
        return self.id

    @abc.abstractproperty
    def id(self):  # pylint: disable=invalid-name
        """
        Return the ID in the DB.
        """

    @abc.abstractproperty
    def enabled(self):
        """
        Is the computer enabled for this user?

        :rtype: bool
        """

    @enabled.setter
    def enabled(self, value):
        """
        Set the enabled state for the computer

        :return: Boolean
        """

    @abc.abstractproperty
    def computer(self):
        """
        The computer that this authinfo relates to

        :return: The corresponding computer
        :rtype: :class:`aiida.orm.Computer`
        """

    @abc.abstractproperty
    def user(self):
        """
        The user that this authinfo relates to

        :return: The corresponding user
        :rtype: :class:`aiida.orm.User`
        """

    @abc.abstractproperty
    def is_stored(self):
        """
        Is it already stored or not?

        :return: Boolean
        :rtype: bool
        """

    @abc.abstractmethod
    def get_auth_params(self):
        """
        Get the dictionary of auth_params

        :return: a dictionary
        """

    @abc.abstractmethod
    def set_auth_params(self, auth_params):
        """
        Set the dictionary of auth_params

        :param auth_params: a dictionary with the new auth_params
        """

    @abc.abstractmethod
    def get_metadata(self):
        """
        Get the metadata dictionary

        :return: a dictionary
        """

    @abc.abstractmethod
    def set_metadata(self, metadata):
        """
        Replace the metadata dictionary in the DB with the provided dictionary
        """

    def get_property(self, name):
        try:
            return self.get_metadata()[name]
        except KeyError:
            raise ValueError('Unknown property: {}'.format(name))

    def set_property(self, name, value):
        metadata = self.get_metadata()
        metadata[name] = value
        self.set_metadata(metadata)


@six.add_metaclass(abc.ABCMeta)
class BackendAuthInfoCollection(backends.BackendCollection[BackendAuthInfo]):
    """The collection of AuthInfo entries."""

    ENTITY_CLASS = BackendAuthInfo

    @abc.abstractmethod
    def delete(self, authinfo_id):
        """
        Remove an AuthInfo from the collection with the given id
        :param authinfo_id: The ID of the authinfo to delete
        """

    @abc.abstractmethod
    def get(self, computer, user):
        """
        Return a AuthInfo given a computer and a user

        :param computer: a Computer instance
        :param user: a User instance
        :return: a AuthInfo object associated to the given computer and user
        :raise aiida.common.NotExistent: if the user is not configured to use computer
        :raise sqlalchemy.orm.exc.MultipleResultsFound: if the user is configured
            more than once to use the computer! Should never happen
        """



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the timezone utility module."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import datetime
import unittest
import time

import aiida.common.timezone as timezone


class TimezoneTest(unittest.TestCase):
    """Tests for the timezone utility module."""

    def test_timezone_now(self):
        """Test timezone.now function."""
        delta = datetime.timedelta(minutes=1)
        ref = timezone.now()
        from_tz = timezone.make_aware(datetime.datetime.fromtimestamp(time.time()))
        self.assertLessEqual(from_tz, ref + delta)
        self.assertGreaterEqual(from_tz, ref - delta)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import six
from six.moves import zip

from .data import Data
from .structure import Site as site_class
from aiida.common.exceptions import ValidationError, InputValidationError

__all__ = ('OrbitalData',)


class OrbitalData(Data):
    """
    Used for storing collections of orbitals, as well as
    providing methods for accessing them internally.
    """

    def clear_orbitals(self):
        """
        Remove all orbitals that were added to the class
        Cannot work if OrbitalData has been already stored
        """
        self.set_attribute('orbital_dicts', [])

    def _get_orbital_class_from_orbital_dict(self, orbital_dict):
        """
        Gets the orbital class from the orbital dictionary stored in DB

        :param orbital_dict: orbital dictionary associated with the orbital
        :return: an Orbital produced using the module_name
        """
        from aiida.plugins import OrbitalFactory

        try:
            module_name = orbital_dict['module_name']
        except KeyError:
            raise ValidationError("No valid module name found in orbital")
        return OrbitalFactory(module_name)

    @staticmethod
    def _prep_orbital_dict_keys_from_site(site):
        """
        Prepares the position from an input site.

        :param site: a site of site class
        :return out_dict: a dictionary of attributes parsed from the site
                          (currently only position)
        """
        if not isinstance(site, site_class):
            raise InputValidationError('Provided input was not a site')

        out_dict = {}
        try:
            position = site.position
            out_dict.update({'position': position})
        except AttributeError:
            raise InputValidationError('site did not have a position!')
        return out_dict

    def get_orbitals(self, with_tags=False, **kwargs):
        """
        Returns all orbitals by default. If a site is provided, returns
        all orbitals cooresponding to the location of that site, additional
        arguments may be provided, which act as filters on the retrieved
        orbitals.

        :param site: if provided, returns all orbitals with position of site
        :param with_tags: if provided returns all tags stored
        :kwargs: attributes than can filter the set of returned orbitals
        :return list_of_outputs: a list of orbitals and also tags if
                                 with_tags was set to True
        """

        import copy
        orbital_dicts = copy.deepcopy(self.get_attribute('orbital_dicts', None))
        if orbital_dicts is None:
            raise AttributeError('Orbitals must be set before')

        filter_dict = {}
        filter_dict.update(kwargs)
        # prevents KeyError from occuring
        orbital_dicts = [x for x in orbital_dicts if all([y in x for y in filter_dict])]
        orbital_dicts = [x for x in orbital_dicts if
                         all([x[y] == filter_dict[y] for y in filter_dict])]

        list_of_outputs = []
        for orbital_dict in orbital_dicts:
            OrbitalClass = self._get_orbital_class_from_orbital_dict(
                            orbital_dict)
            orbital = OrbitalClass()
            try:
                orbital.set_orbital_dict(orbital_dict)
            except ValidationError:
                raise ValueError("Could not reconstruct orbital from data")
            list_of_outputs.append(orbital)
        if with_tags:
            tags = copy.deepcopy(self.get_attribute('tags', None))
            list_of_outputs.append(tags)
        return list_of_outputs

    def set_orbitals(self, orbital, tag=None):
        """
        Sets the orbitals into the database. Uses the orbital's inherent
        set_orbital_dict method to generate a orbital dict string at is stored
        along with the tags, if provided.

        :param orbital: an orbital or list of orbitals to be set
        :param tag: a list of strings must be of length orbital
        """
        from aiida.tools.data.orbital import Orbital

        # convert everything to lists
        if not isinstance(orbital, list):
            orbital = [orbital]
        if not isinstance(tag, list):
            if tag is None:
                tag = ['']*len(orbital)
            else:
                tag = [tag]
        # check if list length matches
        if len(tag) != len(orbital):
            raise ValueError()
        for input_name, this_input, kind in [['orbital', orbital, Orbital],
                                            ['tag', tag, six.string_types]]:
            if not isinstance(this_input, (list, tuple)):
                raise ValueError
            if any([True for _ in this_input if not isinstance(_, kind)]):
                raise ValueError("{} must be a list of {}"
                                 "".format(input_name, kind))
        list_of_tags_to_be_stored = []
        list_of_orbitaldicts_to_be_stored = []
        for this_projection, this_tag in zip(orbital, tag):
            orbital_dict = this_projection.get_orbital_dict()
            OrbitalClass = self._get_orbital_class_from_orbital_dict(
                           orbital_dict)
            test_orbital = OrbitalClass()
            try:
                test_orbital.set_orbital_dict(orbital_dict)
            except ValidationError:
                raise ValueError("The orbital of tag {} did not"
                                 "pass the validation test".format(this_tag))
            orbital_dict = test_orbital.get_orbital_dict()
            list_of_tags_to_be_stored.append(this_tag)
            list_of_orbitaldicts_to_be_stored.append(orbital_dict)
        self.set_attribute('orbital_dicts', list_of_orbitaldicts_to_be_stored)
        if tag is not None:
            self.set_attribute('tags', list_of_tags_to_be_stored)

##########################################################################
#     Here are some ideas for potential future convenience methods
#########################################################################
#     def set_projection_on_site(self, orbital, site, tag=None):
#         """
#         Sets a orbital on a site
#         We prepare the description dictionary, using information `parsed`
#         from the site.
#         """
#         diffusivity = from_site_guess_diffusivity(site) # or 1.
#         position = site.position
#         description = {'somedictionary of the above':''}
#         self.set_projection(orbital=orbital, description=description)

#     def delete_projections_by_attribute(self, selection_attributes):
#         """
#         Deletes all projections whose internal attributes correspond to the
#         selection_attributes
#         """
#         raise NotImplementedError
#
#     def modify_projections(self, key_attributes_to_select_projections, attributes_to_be_modified):
#         """
#         Modifies the projections, as selected by the key_attributes.
#         Overwrites attributes inside these projections, to values stored
#         in attributes_to_be_modified
#         """
#
#    def set_realhydrogenorbitals_from_structure(self, structure, pseudo_family=None):
#        raise NotImplementedError




# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Class to backup an AiiDA instance profile."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import datetime
import io
import logging
import os
import shutil
import stat
import sys

from aiida.backends.profile import BACKEND_DJANGO
from aiida.backends.profile import BACKEND_SQLA
from aiida.backends.utils import load_dbenv, is_dbenv_loaded
from aiida.common import json
from aiida.manage.backup.backup_base import AbstractBackup, BackupError
from aiida.manage.configuration.settings import AIIDA_CONFIG_FOLDER

from . import backup_utils as utils

if not is_dbenv_loaded():
    load_dbenv()

from aiida.backends.settings import BACKEND, AIIDADB_PROFILE  # pylint: disable=wrong-import-order,wrong-import-position


class BackupSetup(object):  # pylint: disable=useless-object-inheritance
    """
    This class setups the main backup script related information & files like::

        - the backup parameter file. It also allows the user to set it up by answering questions.
        - the backup folders.
        - the script that initiates the backup.
    """

    def __init__(self):
        # The backup directory names
        self._conf_backup_folder_rel = "backup_{}".format(AIIDADB_PROFILE)
        self._file_backup_folder_rel = "backup_dest"

        # The backup configuration file (& template) names
        self._backup_info_filename = "backup_info.json"
        self._backup_info_tmpl_filename = "backup_info.json.tmpl"

        # The name of the script that initiates the backup
        self._script_filename = "start_backup.py"

        # Configuring the logging
        logging.basicConfig(format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')

        # The logger of the backup script
        self._logger = logging.getLogger("aiida_backup_setup")

    @staticmethod
    def construct_backup_variables(file_backup_folder_abs):
        """Construct backup variables."""
        backup_variables = {}

        # Setting the oldest backup timestamp
        oldest_object_bk = utils.ask_question(
            "Please provide the oldest backup timestamp "
            "(e.g. 2014-07-18 13:54:53.688484+00:00). ", datetime.datetime, True)

        if oldest_object_bk is None:
            backup_variables[AbstractBackup.OLDEST_OBJECT_BK_KEY] = None
        else:
            backup_variables[AbstractBackup.OLDEST_OBJECT_BK_KEY] = str(oldest_object_bk)

        # Setting the backup directory
        backup_variables[AbstractBackup.BACKUP_DIR_KEY] = file_backup_folder_abs

        # Setting the days_to_backup
        backup_variables[AbstractBackup.DAYS_TO_BACKUP_KEY] = utils.ask_question(
            "Please provide the number of days to backup.", int, True)

        # Setting the end date
        end_date_of_backup_key = utils.ask_question(
            "Please provide the end date of the backup " + "(e.g. 2014-07-18 13:54:53.688484+00:00).",
            datetime.datetime, True)
        if end_date_of_backup_key is None:
            backup_variables[AbstractBackup.END_DATE_OF_BACKUP_KEY] = None
        else:
            backup_variables[AbstractBackup.END_DATE_OF_BACKUP_KEY] = str(end_date_of_backup_key)

        # Setting the backup periodicity
        backup_variables[AbstractBackup.PERIODICITY_KEY] = utils.ask_question("Please periodicity (in days).", int,
                                                                              False)

        # Setting the backup threshold
        backup_variables[AbstractBackup.BACKUP_LENGTH_THRESHOLD_KEY] = utils.ask_question(
            "Please provide the backup threshold (in hours).", int, False)

        return backup_variables

    def create_dir(self, question, dir_path):
        """Create the directories for the backup folder and return its path."""
        final_path = utils.query_string(question, dir_path)

        if not os.path.exists(final_path):
            if utils.query_yes_no("The path {} doesn't exist. Should it be created?".format(final_path), "yes"):
                try:
                    os.makedirs(final_path)
                except OSError:
                    self._logger.error("Error creating the path %s.", final_path)
                    raise
        return final_path

    @staticmethod
    def print_info():
        """Write a string with information to stdout."""
        info_str = \
"""Variables to set up in the JSON file
------------------------------------

 * ``periodicity`` (in days): The backup runs periodically for a number of days
   defined in the periodicity variable. The purpose of this variable is to limit
   the backup to run only on a few number of days and therefore to limit the
   number of files that are backed up at every round. e.g. ``"periodicity": 2``
   Example: if you have files in the AiiDA repositories created in the past 30
   days, and periodicity is 15, the first run will backup the files of the first
   15 days; a second run of the script will backup the next 15 days, completing
   the backup (if it is run within the same day). Further runs will only backup
   newer files, if they are created.

 * ``oldest_object_backedup`` (timestamp or null): This is the timestamp of the
   oldest object that was backed up. If you are not aware of this value or if it
   is the first time that you start a backup up for this repository, then set
   this value to ``null``. Then the script will search the creation date of the
   oldest node object in the database and it will start
   the backup from that date. E.g. ``"oldest_object_backedup":
   "2015-07-20 11:13:08.145804+02:00"``

 * ``end_date_of_backup``: If set, the backup script will backup files that
   have a modification date until the value specified by this variable. If not
   set, the ending of the backup will be set by the following variable
   (``days_to_backup``) which specifies how many days to backup from the start
   of the backup. If none of these variables are set (``end_date_of_backup``
   and ``days_to_backup``), then the end date of backup is set to the current
   date. E.g. ``"end_date_of_backup": null`` or ``"end_date_of_backup":
   "2015-07-20 11:13:08.145804+02:00"``


 * ``days_to_backup``: If set, you specify how many days you will backup from
   the starting date of your backup. If it set to ``null`` and also
   ``end_date_of_backup`` is set to ``null``, then the end date of the backup
   is set to the current date. You can not set ``days_to_backup``
   & ``end_date_of_backup`` at the same time (it will lead to an error).
   E.g. ``"days_to_backup": null`` or ``"days_to_backup": 5``

 * ``backup_length_threshold`` (in hours): The backup script runs in rounds and
   on every round it backs-up a number of days that are controlled primarily by
   ``periodicity`` and also by ``end_date_of_backup`` / ``days_to_backup``,
   for the last backup round. The ``backup_length_threshold`` specifies the
   lowest acceptable round length. This is important for the end of the backup.

 * ``backup_dir``: The destination directory of the backup. e.g.
   ``"backup_dir": "/scratch/aiida_user/backup_script_dest"``
"""
        sys.stdout.write(info_str)

    def run(self):
        """Run the backup."""
        conf_backup_folder_abs = self.create_dir(
            "Please provide the backup folder by providing the full path.",
            os.path.join(os.path.expanduser(AIIDA_CONFIG_FOLDER), self._conf_backup_folder_rel))

        file_backup_folder_abs = self.create_dir(
            "Please provide the destination folder of the backup (normally in "
            "the previously provided backup folder).", os.path.join(conf_backup_folder_abs,
                                                                    self._file_backup_folder_rel))

        # The template backup configuration file
        template_conf_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), self._backup_info_tmpl_filename)

        # Copy the sample configuration file to the backup folder
        try:
            shutil.copy(template_conf_path, conf_backup_folder_abs)
        except Exception:
            self._logger.error("Error copying the file %s to the directory %s", template_conf_path,
                               conf_backup_folder_abs)
            raise

        if utils.query_yes_no(
                "A sample configuration file was copied to {}. "
                "Would you like to ".format(conf_backup_folder_abs) + "see the configuration parameters explanation?",
                default="yes"):
            self.print_info()

        # Construct the path to the backup configuration file
        final_conf_filepath = os.path.join(conf_backup_folder_abs, self._backup_info_filename)

        # If the backup parameters are configured now
        if utils.query_yes_no("Would you like to configure the backup " + "configuration file now?", default="yes"):

            # Ask questions to properly setup the backup variables
            backup_variables = self.construct_backup_variables(file_backup_folder_abs)

            with io.open(final_conf_filepath, 'wb') as backup_info_file:
                json.dump(backup_variables, backup_info_file)
        # If the backup parameters are configured manually
        else:
            sys.stdout.write("Please rename the file {} ".format(self._backup_info_tmpl_filename) +
                             "found in {} to ".format(conf_backup_folder_abs) +
                             "{} and ".format(self._backup_info_filename) +
                             "change the backup parameters accordingly.\n")
            sys.stdout.write("Please adapt the startup script accordingly to point to the " +
                             "correct backup configuration file. For the moment, it points " +
                             "to {}\n".format(os.path.join(conf_backup_folder_abs, self._backup_info_filename)))

        # The contents of the startup script
        if BACKEND == BACKEND_DJANGO:
            backup_import = ("from aiida.manage.backup.backup_django import Backup")
        elif BACKEND == BACKEND_SQLA:
            backup_import = ("from aiida.manage.backup.backup_sqlalchemy import Backup")
        else:
            raise BackupError("Following backend is unknown: {}".format(BACKEND))

        script_content = \
u"""#!/usr/bin/env python
import logging
from aiida.backends.utils import load_dbenv, is_dbenv_loaded

if not is_dbenv_loaded():
    load_dbenv(profile="{}")

{}

# Create the backup instance
backup_inst = Backup(backup_info_filepath="{}", additional_back_time_mins = 2)

# Define the backup logging level
backup_inst._logger.setLevel(logging.INFO)

# Start the backup
backup_inst.run()
""".format(AIIDADB_PROFILE, backup_import, final_conf_filepath)

        # Script full path
        script_path = os.path.join(conf_backup_folder_abs, self._script_filename)

        # Write the contents to the script
        with io.open(script_path, 'w', encoding='utf8') as script_file:
            script_file.write(script_content)

        # Set the right permissions
        try:
            statistics = os.stat(script_path)
            os.chmod(script_path, statistics.st_mode | stat.S_IEXEC)
        except OSError:
            self._logger.error("Problem setting the right permissions to the script %s.", script_path)
            raise

        sys.stdout.write("Backup setup completed.")


if __name__ == '__main__':
    BackupSetup().run()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Translator for calculation node
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import os

from aiida.restapi.translator.node import NodeTranslator
from aiida.restapi.common.exceptions import RestInputValidationError


class CalculationTranslator(NodeTranslator):
    """
    Translator relative to resource 'calculations' and aiida class Calculation
    """

    # A label associated to the present class (coincides with the resource name)
    __label__ = "calculations"
    # The AiiDA class one-to-one associated to the present class
    from aiida.orm import CalcJobNode
    _aiida_class = CalcJobNode
    # The string name of the AiiDA class
    _aiida_type = "process.calculation.calcjob.CalcJobNode"
    # The string associated to the AiiDA class in the query builder lexicon
    _qb_type = _aiida_type + '.'

    _result_type = __label__

    def __init__(self, **kwargs):
        """
        Initialise the parameters.
        Create the basic query_help
        """
        # basic query_help object
        super(CalculationTranslator, self).__init__(Class=self.__class__, **kwargs)

        ## calculation schema
        # All the values from column_order must present in additional info dict
        # Note: final schema will contain details for only the fields present in column order
        self._schema_projections = {
            "column_order": [
                "id", "label", "node_type", "ctime", "mtime", "uuid", "user_id", "user_email", "attributes.state",
                "attributes", "extras"
            ],
            "additional_info": {
                "id": {
                    "is_display": True
                },
                "label": {
                    "is_display": False
                },
                "node_type": {
                    "is_display": True
                },
                "ctime": {
                    "is_display": True
                },
                "mtime": {
                    "is_display": True
                },
                "uuid": {
                    "is_display": False
                },
                "user_id": {
                    "is_display": False
                },
                "user_email": {
                    "is_display": True
                },
                "attributes.state": {
                    "is_display": True
                },
                "attributes": {
                    "is_display": False
                },
                "extras": {
                    "is_display": False
                }
            }
        }

    @staticmethod
    def get_files_list(dirobj, files=None, prefix=None):
        """
        Get list of all files present in given directory.

        :param dirobj: Directory in which files will be searched
        :param files: list of files if any
        :param prefix: file name prefix if any
        :return:
        """
        if files is None:
            files = []
        if prefix is None:
            prefix = []
        for fname in dirobj.get_content_list():
            if dirobj.isfile(fname):
                filename = os.path.join(*(prefix + [fname]))
                files.append(filename)
            elif dirobj.isdir(fname):
                CalculationTranslator.get_files_list(dirobj.get_subfolder(fname), files, prefix + [fname])
        return files

    @staticmethod
    def get_retrieved_inputs(node, filename=None, rtype=None):
        """
        Get the submitted input files for job calculation
        :param node: aiida node
        :return: the retrieved input files for job calculation
        """
        if node.node_type.startswith("process.calculation.calcjob"):

            input_folder = node._raw_input_folder  # pylint: disable=protected-access

            if filename is not None:
                response = {}

                if rtype is None:
                    rtype = "download"

                if rtype == "download":
                    try:
                        content = NodeTranslator.get_file_content(input_folder, filename)
                    except IOError:
                        error = "Error in getting {} content".format(filename)
                        raise RestInputValidationError(error)

                    response["status"] = 200
                    response["data"] = content
                    response["filename"] = filename.replace("/", "_")

                else:
                    raise RestInputValidationError("rtype is not supported")

                return response

            # if filename is not provided, return list of all retrieved files
            retrieved = CalculationTranslator.get_files_list(input_folder)
            return retrieved

        return []

    @staticmethod
    def get_retrieved_outputs(node, filename=None, rtype=None):
        """
        Get the retrieved output files for job calculation
        :param node: aiida node
        :return: the retrieved output files for job calculation
        """
        if node.node_type.startswith("process.calculation.calcjob"):

            retrieved_folder = node.outputs.retrieved
            response = {}

            if retrieved_folder is None:
                response["status"] = 200
                response["data"] = "This node does not have retrieved folder"
                return response

            output_folder = retrieved_folder._get_folder_pathsubfolder  # pylint: disable=protected-access

            if filename is not None:

                if rtype is None:
                    rtype = "download"

                if rtype == "download":
                    try:
                        content = NodeTranslator.get_file_content(output_folder, filename)
                    except IOError:
                        error = "Error in getting {} content".format(filename)
                        raise RestInputValidationError(error)

                    response["status"] = 200
                    response["data"] = content
                    response["filename"] = filename.replace("/", "_")

                else:
                    raise RestInputValidationError("rtype is not supported")

                return response

            # if filename is not provided, return list of all retrieved files
            retrieved = CalculationTranslator.get_files_list(output_folder)
            return retrieved

        return []
